# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from typing import Dict, List, Set, Tuple, Type

import tensorflow as tf
from tensorflow.core.framework.node_def_pb2 import NodeDef
from tensorflow.python.framework.convert_to_constants import convert_variables_to_constants_v2

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.definitions import NNCFGraphNodeType
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.graph.layer_attributes import ConvertDtypeLayerAttributes
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.layer_attributes import Dtype
from nncf.common.graph.layer_attributes import LinearLayerAttributes
from nncf.common.graph.layer_attributes import MultipleInputLayerAttributes
from nncf.common.graph.layer_attributes import MultipleOutputLayerAttributes
from nncf.common.graph.layer_attributes import PermuteLayerAttributes
from nncf.common.graph.layer_attributes import ReshapeLayerAttributes
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.operator_metatypes import OutputNoopMetatype
from nncf.common.graph.utils import get_concat_axis
from nncf.common.graph.utils import get_split_axis
from nncf.common.logging import nncf_logger
from nncf.tensorflow.graph.metatypes import keras_layers as layer_metatypes
from nncf.tensorflow.graph.metatypes.common import CAST_METATYPES
from nncf.tensorflow.graph.metatypes.common import DECONV_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import DEPTHWISE_CONV_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import DIMENSION_PERMUTATION_METATYPES
from nncf.tensorflow.graph.metatypes.common import GENERAL_CONV_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import (
    LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_CONCAT_INPUTS,
)
from nncf.tensorflow.graph.metatypes.common import LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_OUTPUTS
from nncf.tensorflow.graph.metatypes.common import LINEAR_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import RESHAPE_METATYPES
from nncf.tensorflow.graph.metatypes.matcher import get_keras_layer_metatype
from nncf.tensorflow.graph.metatypes.matcher import get_op_metatype
from nncf.tensorflow.graph.metatypes.tf_ops import WEIGHTABLE_TF_OP_METATYPES
from nncf.tensorflow.graph.utils import get_shared_node_name
from nncf.tensorflow.graph.utils import is_builtin_layer
from nncf.tensorflow.graph.utils import is_functional_model
from nncf.tensorflow.graph.utils import is_sequential_model
from nncf.tensorflow.graph.utils import reformat_inbound_nodes_for_oplambda
from nncf.tensorflow.graph.utils import unwrap_layer
from nncf.tensorflow.layers.data_layout import get_input_channel_axis
from nncf.tensorflow.tf_internals import KerasTensor

PREFIX_AUXILIARY_OUTPUT_NODE = "output"


class TFLayerInfo:
    def __init__(self, layer_name: str, instance_idx: int):
        self.layer_name = layer_name
        self.instance_idx = instance_idx


def convert_keras_model_to_nncf_graph(model: tf.keras.Model) -> NNCFGraph:
    """
    Convert Keras model graph to the NNCFGraph

    :param model: Keras model
    :return: NNCFGraph
    """
    converter = TFModelConverterFactory.create(model)
    return converter.convert()


class CustomLayerNodeInfo:
    """
    Describes a future NNCFGraph node corresponding to a TF operation inside a custom layer.
    """

    def __init__(
        self,
        graphdef_node_name: str,
        custom_layer_name: str,
        target_node_name: NNCFNodeName,
        node_type: str,
        node_metatype: Type[OperatorMetatype],
        weight_node_name: str,
        dtype: Dtype,
    ):
        """
        :param graphdef_node_name: The name of the TF operation node in the GraphDef representation
          of the custom layer.
        :param custom_layer_name: The name of the custom layer in the model that hosts it.
        :param target_node_name: The name of the future NNCFNode that will be present in NNCFGraph
          to represent this TF operation in the corresponding custom layer.
        :param node_type: The TF-specific string describing the type of the TF operation.
        :param node_metatype: The metatype to be associated with this operation in the NNCFGraph
        :param weight_node_name: The name of the NNCFGraph node that corresponds to the weight of the operation.
        :param dtype: The dtype of the operation.
        """
        self.graphdef_node_name = graphdef_node_name
        self.custom_layer_name = custom_layer_name
        self.target_node_name = target_node_name
        self.node_type = node_type
        self.node_metatype = node_metatype
        self.weight_node_name = weight_node_name
        self.dtype = dtype


class CustomLayerEdgeInfo:
    """
    Describes an edge inside a custom layer graph representation in TF.
    """

    def __init__(self, tensor_shape: List[int], input_port_id: int, output_port_id: int, dtype: Dtype):
        """
        :param tensor_shape: The shape of the associated tensor.
        :param input_port_id: The ID of the input tensor among the "to_node" inputs.
        :param output_port_id: The ID of the output tensor among the "from_node" outputs.
        :param dtype: The dtype of the tensor.
        """
        self.tensor_shape = tensor_shape
        self.input_port_id = input_port_id
        self.output_port_id = output_port_id
        self.dtype = dtype


class CustomLayerInfo:
    """
    Describes a custom layer in the TF model from the conversion standpoint.
    """

    def __init__(self):
        self.requires_inputs_from_nodes: Set[str] = set()
        self.gives_outputs_from_nodes: Set[str] = set()
        self.shared_weight_node_names_vs_weighted_op_node_names: Dict[str, Set[str]] = defaultdict(set)
        self.node_infos: Dict[str, CustomLayerNodeInfo] = {}
        self.edge_infos: Dict[Tuple[str, str], CustomLayerEdgeInfo] = {}
        self.graphdef_node_name_to_pretty_node_name: Dict[str, str] = {}


class TFModelConverter(ABC):
    """
    Abstract class which describes the interface needed to convert
    the Keras model to the `NNCFGraph` object.
    """

    @abstractmethod
    def convert(self) -> NNCFGraph:
        """
        Converts the Keras model to the `NNCFGraph` object.

        :return: The `NNCFGraph` object that represents the Keras model
            for compression algorithms.
        """


class BaseFunctionalSequentialConverter(TFModelConverter):
    """
    A base class for the `FunctionalConverter` and `SequentialConverter` classes.
    Contains the implementation of the common methods.
    """

    def __init__(self, model: tf.keras.Model):
        self._model = model
        self._node_info: Dict[str, Dict] = {}
        self._custom_layer_infos = self._collect_custom_layer_infos(self._model)
        self._nncf_node_names_vs_custom_layer_name: Dict[NNCFNodeName, str] = {}

    @staticmethod
    def _get_type_spec(tensor):
        if isinstance(tensor, KerasTensor):
            return tensor.type_spec
        return tf.TensorSpec.from_tensor(tensor)

    def _collect_custom_layer_infos(
        self, model: tf.keras.Model, use_graph_var_names: bool = False
    ) -> Dict[str, CustomLayerInfo]:
        custom_layers = BaseFunctionalSequentialConverter.get_custom_layers(model)
        retval = {}
        for layer_name, layer in custom_layers.items():
            layer_input_spec = (
                [self._get_type_spec(tensor) for tensor in layer.input]
                if isinstance(layer.input, list)
                else self._get_type_spec(layer.input)
            )

            # TODO (vshampor) : Use the custom layer's inbound_nodes/outbound_nodes to determine what edges
            #  should connect it to the rest of the graph. Currently the custom layer
            #  subgraph will be present in the NNCFGraph after conversion, which is useful
            #  for purposes of weight modification target point creation and usage,
            #  but the subgraph won't be connected to the rest of the graph; in the main graph
            #  component, the custom layer will still be represented by a single node
            concr_fn = tf.function(layer).get_concrete_function(layer_input_spec, training=False)
            wrapped_function = convert_variables_to_constants_v2(concr_fn, lower_control_flow=False)

            graphdef_nodes = wrapped_function.graph.as_graph_def().node
            graphdef_name_to_layer_var_map = (
                {}
                if use_graph_var_names
                else BaseFunctionalSequentialConverter._get_graphdef_name_to_layer_var_map(concr_fn)
            )
            nodes = {graphdef_name_to_layer_var_map.get(node.name, node.name): node for node in graphdef_nodes}
            graphdef_node_name_vs_node = {node.name: node for node in graphdef_nodes}

            custom_layer_info = CustomLayerInfo()
            for pretty_node_name, node in nodes.items():
                custom_layer_info.graphdef_node_name_to_pretty_node_name[node.name] = pretty_node_name

            for pretty_node_name, node in nodes.items():
                weight_node_name = None
                metatype = get_op_metatype(node.op)
                if metatype in WEIGHTABLE_TF_OP_METATYPES:
                    graphdef_weight_node_name = self._get_graphdef_node_name_for_custom_layer_node_weight(
                        node, graphdef_node_name_vs_node
                    )
                    if graphdef_weight_node_name in graphdef_name_to_layer_var_map:
                        weight_node_name = graphdef_name_to_layer_var_map[graphdef_weight_node_name]
                    else:
                        nncf_logger.warning(
                            f"Could not associate a weighted custom layer node {pretty_node_name} "
                            f"with a weight attribute of the custom layer - the corresponding weight "
                            f"will not be compressed! Make sure that the corresponding custom layer "
                            f"weight has a name."
                        )

                custom_layer_info.node_infos[pretty_node_name] = CustomLayerNodeInfo(
                    graphdef_node_name=node.name,
                    custom_layer_name=layer_name,
                    target_node_name=pretty_node_name,
                    node_type=node.op,
                    node_metatype=get_op_metatype(node.op),
                    weight_node_name=weight_node_name,
                    dtype=Dtype.FLOAT if node.attr["dtype"].type == 1 else Dtype.INTEGER,
                )

                custom_layer_info.shared_weight_node_names_vs_weighted_op_node_names[weight_node_name].add(
                    pretty_node_name
                )

                for idx, input_graphdef_node_name_and_output_port_str in enumerate(node.input):
                    if "^" in input_graphdef_node_name_and_output_port_str:
                        continue  # Skip control_inputs
                    splits = input_graphdef_node_name_and_output_port_str.split(":")
                    if len(splits) == 1:
                        input_graphdef_node_name = splits[0]
                        output_port_id = 0
                    elif len(splits) == 2:
                        input_graphdef_node_name = splits[0]
                        output_port_id = int(splits[1])
                    else:
                        raise nncf.InternalError("Could not parse NodeDef's input field!")

                    pretty_input_node_name = custom_layer_info.graphdef_node_name_to_pretty_node_name[
                        input_graphdef_node_name
                    ]

                    # TODO (vshampor): add proper tensor_shape, will probably involve
                    #                  running as_graph_def(add_shapes=True)
                    custom_layer_info.edge_infos[(pretty_input_node_name, pretty_node_name)] = CustomLayerEdgeInfo(
                        tensor_shape=None,
                        input_port_id=idx,
                        output_port_id=output_port_id,
                        dtype=custom_layer_info.node_infos[pretty_node_name].dtype,
                    )
                retval[layer_name] = custom_layer_info
        return retval

    def _get_layer_output_dtype(self, layer_config: Dict) -> Dtype:
        if layer_config["class_name"] in ["Functional", "Sequential"]:
            return self._get_layer_output_dtype(layer_config["config"]["layers"][0])

        layer_name = layer_config["config"]["name"]
        if layer_config["class_name"] == "TensorFlowOpLayer":
            layer_name = "tf_op_layer_" + layer_name

        if layer_config["class_name"] == "InputLayer":
            return Dtype.FLOAT

        keras_layer = self._get_layer(layer_name)
        if isinstance(keras_layer.output, (tf.Tensor, KerasTensor)):
            dtype = keras_layer.output.dtype
        else:
            # In case of multiple outputs, assume all outputs have the same type
            dtype = keras_layer.output[0].dtype

        if dtype == tf.int32:
            return Dtype.INTEGER
        return Dtype.FLOAT

    @staticmethod
    def _get_layer_type(layer_config: Dict) -> str:
        if layer_config["class_name"] == "TensorFlowOpLayer":
            return layer_config["config"]["node_def"]["op"]
        if layer_config["class_name"] in ["TFOpLambda", "SlicingOpLambda"]:
            return layer_config["config"]["function"]
        if layer_config["class_name"] == "NNCFWrapper":
            # Return class_name of wrapped layer_config
            return layer_config["config"]["layer"]["class_name"]
        return layer_config["class_name"]

    @staticmethod
    def get_custom_layers(model: tf.keras.Model) -> Dict[str, tf.Module]:
        """
        Returns the mapping of custom layer names in the model vs associated custom layer
        module objects.

        :param model: The model in which to search for custom layers.
        :return: A dict of custom layer names vs custom layer modules.
        """
        custom_layers = {}
        for layer in model.submodules:
            if not is_builtin_layer(layer):
                custom_layers[layer.name] = layer
        return custom_layers

    @staticmethod
    def _get_graphdef_name_to_layer_var_map(concrete_fun) -> Dict[str, str]:
        names_map = {}
        inverse_map = defaultdict(set)
        for layer_var in concrete_fun.variables:
            for value_tensor, graphdef_name in concrete_fun.graph.captures:
                if layer_var.handle is value_tensor:
                    graphdef_name = graphdef_name.name.split(":")[0]
                    layer_var_name = layer_var.name.split(":")[0]
                    inverse_map[layer_var_name].add(graphdef_name)
                    names_map[graphdef_name] = layer_var_name

        for graphdef_names in inverse_map.values():
            if len(graphdef_names) > 1:
                # Name collision - remove all collided entries
                for graphdef_name in graphdef_names:
                    del names_map[graphdef_name]

        return names_map

    @staticmethod
    def _get_graphdef_node_name_for_custom_layer_node_weight(
        weighted_node: NodeDef, graphdef_nodes: Dict[str, NodeDef]
    ) -> str:
        def get_node_name(graphdef_node_name: str):
            return graphdef_node_name.split(":")[0]

        weight_node_name = None
        previous_node_names = [get_node_name(node_input) for node_input in weighted_node.input]
        while previous_node_names:
            weight_node_name = get_node_name(previous_node_names[-1])
            weight_node = graphdef_nodes[weight_node_name]  # TODO (vshampor): how correct is this actually?
            previous_node_names = [get_node_name(node_input) for node_input in weight_node.input]

            # Filter control inputs, whatever these are
            previous_node_names = list(filter(lambda x: "^" not in x, previous_node_names))
        if weight_node_name is None:
            raise nncf.InternalError("Could not find a weight node for a weighted node {}".format(weighted_node.name))
        return weight_node_name

    @staticmethod
    def _prepare_shape(shape) -> List:
        if not isinstance(shape, list):
            return [shape]
        return shape

    def get_layer_info_for_node(self, node_name: NNCFNodeName) -> Tuple[bool, TFLayerInfo]:
        """
        :param node_name: The node name in the converted NNCFGraph
        :return: A flag indicating whether the node corresponds to a custom layer,
          and the additional TF-specific information about the layer underlying the node.
        """
        if node_name in self._nncf_node_names_vs_custom_layer_name:
            is_custom = True
            custom_layer_name = self._nncf_node_names_vs_custom_layer_name[node_name]
            insertion_data = TFLayerInfo(custom_layer_name, instance_idx=0)
        else:
            is_custom = False
            node_tf_data = self._node_info[node_name]
            layer_name = node_tf_data["layer_name"]
            instance_idx = node_tf_data["inbound_node_idx"]
            if instance_idx is None:
                instance_idx = 0

            insertion_data = TFLayerInfo(layer_name, instance_idx=instance_idx)
        return is_custom, insertion_data

    def get_node_names_vs_custom_layer_names(self) -> Dict[NNCFNodeName, str]:
        """
        :return: A mapping of NNCFNode names corresponding to custom layers vs the corresponding
          custom layer name.
        """
        return self._nncf_node_names_vs_custom_layer_name

    def _get_layer(self, layer_name: str) -> tf.keras.layers.Layer:
        try:
            return self._model.get_layer(layer_name)
        except ValueError:
            for layer in self._model.submodules:
                if not isinstance(layer, tf.keras.layers.Layer):
                    continue
                if layer.name == layer_name:
                    return layer

        raise ValueError(f"No such layer: {layer_name}.")

    def _add_custom_layer_subgraph(self, nncf_graph: NNCFGraph, custom_layer_name: str) -> NNCFGraph:
        # TODO (vshampor): filter meaningless ops such as Identity, resource read etc.
        custom_layer_info = self._custom_layer_infos[custom_layer_name]
        node_name_vs_nncf_node_ids: Dict[NNCFNodeName, int] = {}
        for node_info in custom_layer_info.node_infos.values():
            weight_node_name = node_info.weight_node_name
            is_shared = False
            if weight_node_name is not None:
                shared_node_dict = custom_layer_info.shared_weight_node_names_vs_weighted_op_node_names
                is_shared = len(shared_node_dict[weight_node_name]) > 1
            nncf_node = nncf_graph.add_nncf_node(
                node_name=node_info.target_node_name,
                node_type=node_info.node_type,
                node_metatype=node_info.node_metatype,
                # TODO (vshampor): collect layer attributes for custom nodes
                layer_attributes=None,
                layer_name=node_info.weight_node_name,  # sic!
                is_shared=is_shared,
                ignored_algorithms=["quantization"],
            )
            node_name_vs_nncf_node_ids[node_info.target_node_name] = nncf_node.node_id
            self._nncf_node_names_vs_custom_layer_name[node_info.target_node_name] = custom_layer_name
        for edge, edge_data in custom_layer_info.edge_infos.items():
            from_node_name, to_node_name = edge
            from_node_id = node_name_vs_nncf_node_ids[from_node_name]
            to_node_id = node_name_vs_nncf_node_ids[to_node_name]
            nncf_graph.add_edge_between_nncf_nodes(
                from_node_id,
                to_node_id,
                tensor_shape=edge_data.tensor_shape,
                input_port_id=edge_data.input_port_id,
                output_port_id=edge_data.output_port_id,
                dtype=edge_data.dtype,
            )
        return nncf_graph


class TFModelConverterFactory:
    @staticmethod
    def create(model) -> TFModelConverter:
        func_model = is_functional_model(model)
        seq_model = is_sequential_model(model)

        if not func_model and not seq_model:
            RuntimeError("Only sequential or functional models are supported")

        if func_model:
            converter = FunctionalConverter(model)
        else:
            converter = SequentialConverter(model)
        return converter


class FunctionalConverter(BaseFunctionalSequentialConverter):
    """
    Converter for TF models that use the Functional API.
    """

    def __init__(self, model: tf.keras.Model):
        super().__init__(model)
        self._model_config = self._model.get_config()
        self._layer_info: Dict[str, Dict] = {}
        self._collect_layer_information()
        self._layer_name_to_node_names = defaultdict(set)
        self._collect_node_information()
        self._edge_info: Dict[Tuple[str, str], Dict] = {}
        self._collect_edge_information()

    def _collect_layer_information(self):
        for layer_config in self._model_config["layers"]:
            layer_name = layer_config["name"]
            layer_type = self._get_layer_type(layer_config)
            layer_output_dtype = self._get_layer_output_dtype(layer_config)
            data_format = layer_config["config"].get("data_format")
            model_layer = self._get_layer(layer_name)
            layer_metatype = get_keras_layer_metatype(model_layer)
            self._layer_info[layer_name] = {
                "type": layer_type,
                "metatype": layer_metatype,
                "dtype": layer_output_dtype,
                "data_format": data_format,
                "inbound_nodes": layer_config.get("inbound_nodes"),
            }

    def _collect_node_information(self):
        for layer_config in self._model_config["layers"]:
            layer_name = layer_config["name"]
            if layer_name not in self._custom_layer_infos:
                self._add_regular_layer_nodes(layer_config)
            else:
                # TODO (vshampor): Instead of adding a single node for custom layer and an entire
                #  unconnected subgraph along with it, stitch the subgraph into the main graph
                #  properly
                self._add_regular_layer_nodes(layer_config)

    def _add_regular_layer_nodes(self, layer_config: Dict):
        layer_name = layer_config["name"]
        layer = self._get_layer(layer_name)
        if layer_config["inbound_nodes"]:
            instances_count = len(layer_config["inbound_nodes"])
            is_shared = instances_count > 1
            for i in range(instances_count):
                node_name = get_shared_node_name(layer_name, i) if is_shared else layer_name
                input_shapes = [tuple(tensor.shape) for tensor in layer.inbound_nodes[i].keras_inputs]
                output_shapes = self._prepare_shape(layer.inbound_nodes[i].output_shapes)
                self._node_info[node_name] = {
                    "layer_name": layer_name,
                    "target_node_name": layer_name,
                    "inbound_node_idx": i,
                    "input_shapes": input_shapes,
                    "output_shapes": output_shapes,
                }
                self._layer_name_to_node_names[layer_name].add(node_name)
        else:
            node_name = layer_name
            input_shapes = self._prepare_shape(layer.input_shape)
            output_shapes = self._prepare_shape(layer.output_shape)
            self._node_info[node_name] = {
                "layer_name": layer_name,
                "target_node_name": layer_name,
                "inbound_node_idx": None,
                "input_shapes": input_shapes,
                "output_shapes": output_shapes,
            }

    def _is_layer_shared(self, layer_name: str):
        # Only gives valid results if called after collect_node_information()
        return len(self._layer_name_to_node_names[layer_name]) > 1

    def _collect_edge_information(self):
        for layer_config in self._model_config["layers"]:
            layer_name = layer_config["name"]

            inbound_nodes = layer_config["inbound_nodes"]
            if layer_config["class_name"] in ["TFOpLambda", "SlicingOpLambda"]:
                inbound_nodes = reformat_inbound_nodes_for_oplambda(inbound_nodes)

            for layer_instance_idx, inbound_nodes in enumerate(inbound_nodes):
                if self._is_layer_shared(layer_name):
                    node_name = get_shared_node_name(layer_name, layer_instance_idx)
                else:
                    node_name = layer_name
                input_shapes = self._node_info[node_name]["input_shapes"]

                for layer_instance_input_port_id, inbound_node in enumerate(inbound_nodes):
                    producer_layer_name, producer_layer_instance, producer_layer_instance_output_port, _ = inbound_node

                    if self._is_layer_shared(producer_layer_name):
                        producer_node_name = get_shared_node_name(producer_layer_name, producer_layer_instance)
                    else:
                        producer_node_name = producer_layer_name

                    producer_layer_info = self._layer_info[producer_layer_name]
                    dtype = producer_layer_info["dtype"]
                    tensor_shape = input_shapes[layer_instance_input_port_id]

                    edge = (producer_node_name, node_name)
                    self._edge_info[edge] = {
                        "tensor_shape": tensor_shape,
                        "dtype": dtype,
                        "to_node_input_port_id": layer_instance_input_port_id,
                        "from_node_output_port_id": producer_layer_instance_output_port,
                    }

    def convert(self) -> NNCFGraph:
        nncf_graph = NNCFGraph()
        node_name_vs_nncf_node_ids: Dict[str, int] = {}
        output_node_id_vs_model_output_idx: Dict[int, int] = {}

        # Regular nodes
        for node_name, node_info in self._node_info.items():
            layer_name = node_info["layer_name"]
            node_name_vs_nncf_node_ids[layer_name] = []
            layer_info = self._layer_info[layer_name]
            metatype = layer_info["metatype"]
            layer = self._get_layer(layer_name)
            layer_attributes = _get_layer_attributes(metatype, layer)
            is_shared = len(self._layer_name_to_node_names[layer_name]) > 1
            nncf_node = nncf_graph.add_nncf_node(
                node_name=node_name,
                node_type=layer_info["type"],
                node_metatype=metatype,
                layer_attributes=layer_attributes,
                layer_name=layer_name,
                is_shared=is_shared,
            )
            node_name_vs_nncf_node_ids[node_name] = nncf_node.node_id

            if layer in self._model._output_layers:
                output_idx = self._model._output_layers.index(layer)
                output_node_id_vs_model_output_idx[nncf_node.node_id] = output_idx

        # Regular edges
        for edge, edge_info in self._edge_info.items():
            from_node_name, to_node_name = edge
            from_node_id = node_name_vs_nncf_node_ids[from_node_name]
            to_node_id = node_name_vs_nncf_node_ids[to_node_name]
            nncf_graph.add_edge_between_nncf_nodes(
                from_node_id,
                to_node_id,
                tensor_shape=edge_info["tensor_shape"],
                input_port_id=edge_info["to_node_input_port_id"],
                output_port_id=edge_info["from_node_output_port_id"],
                dtype=edge_info["dtype"],
            )

        # Custom nodes and edges
        for custom_layer_name in self._custom_layer_infos:
            nncf_graph = self._add_custom_layer_subgraph(nncf_graph, custom_layer_name)
            # TODO (vshampor): connect the subgraphs with the rest of the graph

        for output_node_id, model_output_idx in output_node_id_vs_model_output_idx.items():
            # Ticket: 56853
            # We won't add an NNCF output auxiliary node for all of the NNCF nodes corresponding to real
            # model output, but only for the nodes that do not serve as a tensor source for any other node.
            # The reason is that current TF capabilities do not allow to insert post-hooks after TF functional model
            # output nodes without changing the name of the corresponding output, which won't be obvious to the user.
            nncf_node = nncf_graph.get_node_by_id(output_node_id)
            if not nncf_graph.get_next_nodes(nncf_node):
                output_aux_node_name = f"{nncf_node.node_name}_{PREFIX_AUXILIARY_OUTPUT_NODE}_{model_output_idx}"
                output_node = nncf_graph.add_nncf_node(
                    node_name=output_aux_node_name,
                    node_type=NNCFGraphNodeType.OUTPUT_NODE,
                    node_metatype=OutputNoopMetatype,
                )
                node_info = self._node_info[nncf_node.node_name]  # works if _node_info keys are identical to node_names
                nncf_graph.add_edge_between_nncf_nodes(
                    nncf_node.node_id,
                    output_node.node_id,
                    tensor_shape=node_info["output_shapes"][0],
                    input_port_id=0,
                    output_port_id=0,
                    dtype=Dtype.FLOAT,
                )

        return nncf_graph


class SequentialConverter(BaseFunctionalSequentialConverter):
    """
    Converter for the TF models using the Sequential API.
    """

    def convert(self) -> NNCFGraph:
        nncf_graph = NNCFGraph()
        producer_layer_id = None
        model_config = self._model.get_config()

        layer_name = None
        for layer_config in model_config["layers"]:
            layer_name = layer_config["config"]["name"]
            if layer_name in self._custom_layer_infos:
                nncf_graph = self._add_custom_layer_subgraph(nncf_graph, layer_name)
                continue
            layer_type = self._get_layer_type(layer_config)
            layer_output_dtype = self._get_layer_output_dtype(layer_config)
            data_format = layer_config["config"].get("data_format")
            model_layer = self._get_layer(layer_name)
            layer_metatype = get_keras_layer_metatype(model_layer)

            attrs = dict(
                type=layer_type,
                dtype=layer_output_dtype,
                metatype=layer_metatype,
                data_format=data_format,
                in_ports=[0],
                out_ports=[0],
                is_shared=False,
            )

            layer_attributes = _get_layer_attributes(layer_metatype, model_layer)
            if layer_attributes is not None:
                attrs.update({NNCFNode.LAYER_ATTRIBUTES: layer_attributes})

            node_name = layer_name
            nncf_node = nncf_graph.add_nncf_node(
                node_name,
                node_type=layer_type,
                node_metatype=layer_metatype,
                layer_attributes=layer_attributes,
                layer_name=layer_name,
                is_shared=False,
            )

            input_shapes = self._prepare_shape(model_layer.input_shape)
            output_shapes = self._prepare_shape(model_layer.output_shape)
            self._node_info[node_name] = {
                "layer_name": layer_name,
                "target_node_name": layer_name,
                "inbound_node_idx": None,
                "input_shapes": input_shapes,
                "output_shapes": output_shapes,
            }

            if producer_layer_id is not None:
                input_shapes = self._prepare_shape(self._model.get_layer(layer_name).input_shape)
                nncf_graph.add_edge_between_nncf_nodes(
                    producer_layer_id,
                    nncf_node.node_id,
                    tensor_shape=input_shapes[0],
                    input_port_id=0,
                    output_port_id=0,
                    dtype=Dtype.FLOAT,
                )  # TODO(vshampor): determine from keras layers
            producer_layer_id = nncf_node.node_id

        if layer_name is not None:
            last_producer_layer_name = layer_name
            last_producer_layer_id = producer_layer_id
            output_model_layer = self._model.get_layer(last_producer_layer_name)
            output_aux_node_name = PREFIX_AUXILIARY_OUTPUT_NODE + "_0"
            output_node = nncf_graph.add_nncf_node(
                node_name=output_aux_node_name,
                node_type=NNCFGraphNodeType.OUTPUT_NODE,
                node_metatype=OutputNoopMetatype,
            )
            nncf_graph.add_edge_between_nncf_nodes(
                last_producer_layer_id,
                output_node.node_id,
                tensor_shape=self._prepare_shape(output_model_layer.output_shape),
                input_port_id=0,
                output_port_id=0,
                dtype=Dtype.FLOAT,
            )

        return nncf_graph


def _get_layer_attributes(
    layer_metatype: Type[OperatorMetatype], model_layer: tf.keras.layers.Layer
) -> BaseLayerAttributes:
    layer_attributes = None
    if layer_metatype in DEPTHWISE_CONV_LAYER_METATYPES:
        layer_attributes = _get_conv_layer_attributes(model_layer, is_depthwise=True)
    elif layer_metatype in GENERAL_CONV_LAYER_METATYPES:
        layer_attributes = _get_conv_layer_attributes(model_layer, is_depthwise=False)
    elif layer_metatype in LINEAR_LAYER_METATYPES:
        layer_attributes = _get_linear_layer_attributes(model_layer)
    elif layer_metatype in LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_CONCAT_INPUTS:
        layer_attributes = _get_multiple_input_layer_attributes(model_layer)
    elif layer_metatype in RESHAPE_METATYPES:
        layer_attributes = _get_reshape_layer_attributes(model_layer)
    elif layer_metatype in DIMENSION_PERMUTATION_METATYPES:
        layer_attributes = _get_permutation_layer_attributes(model_layer, layer_metatype)
    elif layer_metatype in LAYER_METATYPES_AGNOSTIC_TO_DATA_PRECISION_WITH_MULTIPLE_OUTPUTS:
        layer_attributes = _get_multiple_output_layer_attributes(model_layer)
    elif layer_metatype in CAST_METATYPES:
        layer_attributes = _get_cast_layer_attributes(model_layer)

    return layer_attributes


def _get_multiple_input_layer_attributes(layer: tf.keras.layers.Layer) -> MultipleInputLayerAttributes:
    if hasattr(layer, "axis"):
        axis = layer.axis
    else:
        input_shape = layer.input_shape
        output_shape = layer.output_shape
        if not isinstance(output_shape, list):
            output_shape = [output_shape]
        axis = get_concat_axis(input_shape, output_shape)
    return MultipleInputLayerAttributes(axis)


def _get_conv_layer_attributes(layer: tf.keras.layers.Layer, is_depthwise: bool = False) -> ConvolutionLayerAttributes:
    channel_axis = get_input_channel_axis(layer)
    layer_ = unwrap_layer(layer)
    layer_metatype = get_keras_layer_metatype(layer_, determine_subtype=False)
    strides = layer_.strides[0]
    dilations = layer_.dilation_rate
    in_channels = layer.get_input_shape_at(0)[channel_axis]
    out_channels = layer.get_output_shape_at(0)[channel_axis]
    with_bias = layer_.use_bias

    # TF does not deign to properly set the groups attribute on a depthwise layer, and for compatibility
    # with common code the groups attribute of the returned ConvolutionLayerAttribute must be set equal
    # to the in_channels attribute in order for the layer to be detected as depthwise
    groups = layer_.groups if not is_depthwise else in_channels
    kernel_size = layer_.kernel_size

    transpose = layer_metatype in DECONV_LAYER_METATYPES

    return ConvolutionLayerAttributes(
        weight_requires_grad=layer.trainable,
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=kernel_size,
        stride=strides,
        dilations=dilations,
        groups=groups,
        transpose=transpose,
        padding_values=([0, 0, 0, 0]),
        with_bias=with_bias,
    )


def _get_linear_layer_attributes(layer: tf.keras.layers.Layer) -> LinearLayerAttributes:
    channel_axis = get_input_channel_axis(layer)
    in_features = layer.get_input_shape_at(0)[channel_axis]
    out_features = layer.get_output_shape_at(0)[channel_axis]
    with_bias = layer.use_bias
    return LinearLayerAttributes(
        weight_requires_grad=layer.trainable, in_features=in_features, out_features=out_features, with_bias=with_bias
    )


def _get_reshape_layer_attributes(layer: tf.keras.layers.Layer) -> ReshapeLayerAttributes:
    input_shape = layer.input_shape
    output_shape = layer.output_shape
    if isinstance(output_shape, list):
        output_shape = output_shape[0]
    return ReshapeLayerAttributes(input_shape, output_shape)


def _get_permutation_layer_attributes(
    layer: tf.keras.layers.Layer, layer_metatype: OperatorMetatype
) -> PermuteLayerAttributes:
    dims = None
    if hasattr(layer, "dims"):
        dims = list(layer.dims)
    else:
        inbound_nodes = layer.inbound_nodes
        if len(inbound_nodes) > 0 and hasattr(inbound_nodes[0], "call_kwargs"):
            call_kwargs = inbound_nodes[0].call_kwargs
            if "perm" in call_kwargs:
                dims = call_kwargs["perm"]
    assert dims, "failed to parse permute attributes"
    if layer_metatype is layer_metatypes.TFPermuteLayerMetatype:
        # Indexing starts at 1 for Permute.
        # Need a zero dimension for consistent representation of permutation.
        dims.insert(0, 0)
    return PermuteLayerAttributes(dims)


def _get_multiple_output_layer_attributes(layer: tf.keras.layers.Layer) -> MultipleOutputLayerAttributes:
    input_shape = layer.input_shape
    output_shape = layer.output_shape
    chunks = len(output_shape)
    if not isinstance(input_shape, list):
        input_shape = [input_shape]
    axis = get_split_axis(input_shape, output_shape)
    return MultipleOutputLayerAttributes(chunks, axis)


def _get_cast_layer_attributes(layer: tf.keras.layers.Layer) -> ConvertDtypeLayerAttributes:
    src_dtype = layer.input.dtype
    dst_dtype = layer.output.dtype
    return ConvertDtypeLayerAttributes(src_dtype, dst_dtype)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Union

import numpy as np

from nncf.experimental.tensor.functions import linalg
from nncf.experimental.tensor.functions.dispatcher import register_numpy_types


@register_numpy_types(linalg.norm)
def _(
    a: Union[np.ndarray, np.generic],
    ord: Optional[Union[str, float, int]] = None,
    axis: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdims: bool = False,
) -> np.ndarray:
    return np.array(np.linalg.norm(a, ord=ord, axis=axis, keepdims=keepdims))


@register_numpy_types(linalg.cholesky)
def _(a: Union[np.ndarray, np.generic], upper: bool = False) -> np.ndarray:
    lt = np.linalg.cholesky(a)
    if upper:
        return np.conjugate(np.swapaxes(lt, -2, -1))
    return lt


@register_numpy_types(linalg.cholesky_inverse)
def _(a: Union[np.ndarray, np.generic], upper: bool = False) -> np.ndarray:
    c = np.linalg.inv(a)
    ct = np.conjugate(np.swapaxes(c, -2, -1))
    if upper:
        return np.matmul(c, ct)
    return np.matmul(ct, c)


@register_numpy_types(linalg.inv)
def _(a: Union[np.ndarray, np.generic]) -> np.ndarray:
    return np.linalg.inv(a)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np

from nncf.experimental.tensor.definitions import TensorBackend
from nncf.experimental.tensor.definitions import TensorDataType
from nncf.experimental.tensor.definitions import TensorDeviceType
from nncf.experimental.tensor.definitions import TypeInfo
from nncf.experimental.tensor.functions import numeric as numeric
from nncf.experimental.tensor.functions.dispatcher import register_numpy_types

DTYPE_MAP = {
    TensorDataType.float16: np.dtype(np.float16),
    TensorDataType.float32: np.dtype(np.float32),
    TensorDataType.float64: np.dtype(np.float64),
    TensorDataType.int8: np.dtype(np.int8),
    TensorDataType.int32: np.dtype(np.int32),
    TensorDataType.int64: np.dtype(np.int64),
    TensorDataType.uint8: np.dtype(np.uint8),
}

DTYPE_MAP_REV = {v: k for k, v in DTYPE_MAP.items()}


@register_numpy_types(numeric.device)
def _(a: Union[np.ndarray, np.generic]) -> TensorDeviceType:
    return TensorDeviceType.CPU


@register_numpy_types(numeric.backend)
def _(a: Union[np.ndarray, np.generic]) -> TensorBackend:
    return TensorBackend.numpy


@register_numpy_types(numeric.squeeze)
def _(
    a: Union[np.ndarray, np.generic], axis: Optional[Union[int, Tuple[int, ...]]] = None
) -> Union[np.ndarray, np.generic]:
    return np.squeeze(a, axis=axis)


@register_numpy_types(numeric.flatten)
def _(a: Union[np.ndarray, np.generic]) -> np.ndarray:
    return a.flatten()


@register_numpy_types(numeric.max)
def _(
    a: Union[np.ndarray, np.generic], axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: bool = False
) -> np.ndarray:
    return np.array(np.max(a, axis=axis, keepdims=keepdims))


@register_numpy_types(numeric.min)
def _(
    a: Union[np.ndarray, np.generic], axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: bool = False
) -> Union[np.ndarray, np.generic]:
    return np.array(np.min(a, axis=axis, keepdims=keepdims))


@register_numpy_types(numeric.abs)
def _(a: Union[np.ndarray, np.generic]) -> Union[np.ndarray, np.generic]:
    return np.absolute(a)


@register_numpy_types(numeric.astype)
def _(a: Union[np.ndarray, np.generic], dtype: TensorDataType) -> Union[np.ndarray, np.generic]:
    return a.astype(DTYPE_MAP[dtype])


@register_numpy_types(numeric.dtype)
def _(a: Union[np.ndarray, np.generic]) -> TensorDataType:
    return DTYPE_MAP_REV[np.dtype(a.dtype)]


@register_numpy_types(numeric.reshape)
def _(a: Union[np.ndarray, np.generic], shape: Union[int, Tuple[int, ...]]) -> np.ndarray:
    return a.reshape(shape)


@register_numpy_types(numeric.all)
def _(a: Union[np.ndarray, np.generic], axis: Optional[Union[int, Tuple[int, ...]]] = None) -> np.ndarray:
    return np.array(np.all(a, axis=axis))


@register_numpy_types(numeric.allclose)
def _(
    a: Union[np.ndarray, np.generic],
    b: Union[np.ndarray, np.generic, float],
    rtol: float = 1e-05,
    atol: float = 1e-08,
    equal_nan: bool = False,
) -> bool:
    return np.allclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)


@register_numpy_types(numeric.any)
def _(a: Union[np.ndarray, np.generic], axis: Optional[Union[int, Tuple[int, ...]]] = None) -> np.ndarray:
    return np.array(np.any(a, axis=axis))


@register_numpy_types(numeric.count_nonzero)
def _(a: Union[np.ndarray, np.generic], axis: Optional[Union[int, Tuple[int, ...]]] = None) -> np.ndarray:
    return np.array(np.count_nonzero(a, axis=axis))


@register_numpy_types(numeric.isempty)
def _(a: Union[np.ndarray, np.generic]) -> bool:
    return a.size == 0


@register_numpy_types(numeric.isclose)
def _(
    a: Union[np.ndarray, np.generic],
    b: Union[np.ndarray, np.generic, float],
    rtol: float = 1e-05,
    atol: float = 1e-08,
    equal_nan: bool = False,
) -> np.ndarray:
    return np.isclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)


@register_numpy_types(numeric.maximum)
def _(x1: Union[np.ndarray, np.generic], x2: Union[np.ndarray, np.generic, float]) -> np.ndarray:
    return np.maximum(x1, x2)


@register_numpy_types(numeric.minimum)
def _(x1: Union[np.ndarray, np.generic], x2: Union[np.ndarray, np.generic, float]) -> np.ndarray:
    return np.minimum(x1, x2)


@register_numpy_types(numeric.ones_like)
def _(a: Union[np.ndarray, np.generic]) -> np.ndarray:
    return np.ones_like(a)


@register_numpy_types(numeric.where)
def _(
    condition: Union[np.ndarray, np.generic],
    x: Union[np.ndarray, np.generic, float],
    y: Union[np.ndarray, np.generic, float],
) -> np.ndarray:
    return np.where(condition, x, y)


@register_numpy_types(numeric.zeros_like)
def _(a: Union[np.ndarray, np.generic]) -> np.ndarray:
    return np.zeros_like(a)


@register_numpy_types(numeric.stack)
def _(x: Union[np.ndarray, np.generic], axis: int = 0) -> List[np.ndarray]:
    return np.stack(x, axis=axis)


@register_numpy_types(numeric.concatenate)
def _(x: Union[np.ndarray, np.generic], axis: int = 0) -> List[np.ndarray]:
    return np.concatenate(x, axis=axis)


@register_numpy_types(numeric.unstack)
def _(x: Union[np.ndarray, np.generic], axis: int = 0) -> List[np.ndarray]:
    return [np.squeeze(e, axis) for e in np.split(x, x.shape[axis], axis=axis)]


@register_numpy_types(numeric.moveaxis)
def _(a: np.ndarray, source: Union[int, Tuple[int, ...]], destination: Union[int, Tuple[int, ...]]) -> np.ndarray:
    return np.moveaxis(a, source, destination)


@register_numpy_types(numeric.mean)
def _(
    a: Union[np.ndarray, np.generic],
    axis: Union[int, Tuple[int, ...]] = None,
    keepdims: bool = False,
    dtype: Optional[TensorDataType] = None,
) -> np.ndarray:
    dtype = DTYPE_MAP[dtype] if dtype else None
    return np.array(np.mean(a, axis=axis, keepdims=keepdims, dtype=dtype))


@register_numpy_types(numeric.median)
def _(
    a: Union[np.ndarray, np.generic],
    axis: Union[int, Tuple[int, ...]] = None,
    keepdims: bool = False,
) -> np.ndarray:
    return np.array(np.median(a, axis=axis, keepdims=keepdims))


@register_numpy_types(numeric.round)
def _(a: Union[np.ndarray, np.generic], decimals: int = 0) -> np.ndarray:
    return np.round(a, decimals=decimals)


@register_numpy_types(numeric.power)
def _(a: Union[np.ndarray, np.generic], exponent: Union[np.ndarray, float]) -> Union[np.ndarray, np.generic]:
    return np.power(a, exponent)


@register_numpy_types(numeric.quantile)
def _(
    a: Union[np.ndarray, np.generic],
    q: Union[float, List[float]],
    axis: Optional[Union[int, Tuple[int]]] = None,
    keepdims: Optional[bool] = None,
) -> Union[np.ndarray, np.generic]:
    return np.array(np.quantile(a, q=q, axis=axis, keepdims=keepdims))


@register_numpy_types(numeric.percentile)
def _(
    tensor: np.ndarray,
    q: Union[float, List[float]],
    axis: Union[int, Tuple[int, ...], List[int]],
    keepdims: bool = False,
) -> List[Union[np.ndarray, np.generic]]:
    return np.quantile(tensor, q=np.true_divide(np.array(q), 100), axis=axis, keepdims=keepdims)


@register_numpy_types(numeric._binary_op_nowarn)
def _(
    a: Union[np.ndarray, np.generic], b: Union[np.ndarray, np.generic, float], operator_fn: Callable
) -> Union[np.ndarray, np.generic]:
    # Run operator with disabled warning
    with np.errstate(invalid="ignore", divide="ignore"):
        return operator_fn(a, b)


@register_numpy_types(numeric._binary_reverse_op_nowarn)
def _(
    a: Union[np.ndarray, np.generic], b: Union[np.ndarray, np.generic, float], operator_fn: Callable
) -> Union[np.ndarray, np.generic]:
    # Run operator with disabled warning
    with np.errstate(invalid="ignore", divide="ignore"):
        return operator_fn(b, a)


@register_numpy_types(numeric.finfo)
def _(a: np.ndarray) -> TypeInfo:
    ti = np.finfo(a.dtype)
    return TypeInfo(ti.eps, ti.max, ti.min)


@register_numpy_types(numeric.clip)
def _(
    a: Union[np.ndarray, np.generic],
    a_min: Union[np.ndarray, np.generic, float],
    a_max: Union[np.ndarray, np.generic, float],
) -> Union[np.ndarray, np.generic]:
    return np.clip(a, a_min, a_max)


@register_numpy_types(numeric.as_tensor_like)
def _(a: Union[np.ndarray, np.generic], data: Any) -> Union[np.ndarray, np.generic]:
    return np.array(data)


@register_numpy_types(numeric.item)
def _(a: Union[np.ndarray, np.generic]) -> Union[int, float, bool]:
    return a.item()


@register_numpy_types(numeric.sum)
def _(
    a: Union[np.ndarray, np.generic], axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: bool = False
) -> np.ndarray:
    return np.array(np.sum(a, axis=axis, keepdims=keepdims))


@register_numpy_types(numeric.multiply)
def _(x1: Union[np.ndarray, np.generic], x2: Union[np.ndarray, np.generic, float]) -> np.ndarray:
    return np.multiply(x1, x2)


@register_numpy_types(numeric.var)
def _(
    a: Union[np.ndarray, np.generic],
    axis: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdims: bool = False,
    ddof: int = 0,
) -> np.ndarray:
    return np.array(np.var(a, axis=axis, keepdims=keepdims, ddof=ddof))


@register_numpy_types(numeric.size)
def _(a: Union[np.ndarray, np.generic]) -> int:
    return a.size


@register_numpy_types(numeric.matmul)
def _(x1: Union[np.ndarray, np.generic], x2: Union[np.ndarray, np.generic, float]) -> np.ndarray:
    return np.matmul(x1, x2)


@register_numpy_types(numeric.unsqueeze)
def _(
    a: Union[np.ndarray, np.generic], axis: Optional[Union[int, Tuple[int, ...]]] = None
) -> Union[np.ndarray, np.generic]:
    return np.expand_dims(a, axis=axis)


@register_numpy_types(numeric.transpose)
def _(a: Union[np.ndarray, np.generic], axes: Optional[Tuple[int, ...]] = None) -> Union[np.ndarray, np.generic]:
    return np.transpose(a, axes=axes)


@register_numpy_types(numeric.argsort)
def _(
    a: Union[np.ndarray, np.generic], axis: int = -1, descending=False, stable=False
) -> Union[np.ndarray, np.generic]:
    if descending and stable:
        return a.shape[axis] - 1 - np.flip(np.argsort(np.flip(a, axis), axis=axis, kind="stable"), axis)
    if descending and not stable:
        return np.flip(np.argsort(a, axis=axis), axis)
    return np.argsort(a, axis=axis, kind="stable" if stable else None)


@register_numpy_types(numeric.diag)
def _(a: Union[np.ndarray, np.generic], k: int = 0) -> np.ndarray:
    return np.diag(a, k=k)


@register_numpy_types(numeric.logical_or)
def _(x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
    return np.logical_or(x1, x2)


@register_numpy_types(numeric.masked_mean)
def _(
    x: np.ndarray, mask: Optional[np.ndarray], axis: Union[int, Tuple[int, ...], List[int]], keepdims=False
) -> np.ndarray:
    if mask is None:
        return np.mean(x, axis=axis, keepdims=keepdims)
    masked_x = np.ma.array(x, mask=mask)
    result = np.ma.mean(masked_x, axis=axis, keepdims=keepdims)
    if isinstance(result, np.ma.MaskedArray):
        return result.data
    return result


@register_numpy_types(numeric.masked_median)
def _(
    x: np.ndarray, mask: Optional[np.ndarray], axis: Union[int, Tuple[int, ...], List[int]], keepdims=False
) -> np.ndarray:
    if mask is None:
        return np.median(x, axis=axis, keepdims=keepdims)
    masked_x = np.ma.array(x, mask=mask)
    result = np.ma.median(masked_x, axis=axis, keepdims=keepdims)
    if isinstance(result, np.ma.MaskedArray):
        return result.data
    return result


@register_numpy_types(numeric.clone)
def _(a: Union[np.ndarray, np.generic]) -> np.ndarray:
    return a.copy()


@register_numpy_types(numeric.searchsorted)
def _(a: np.ndarray, v: np.ndarray, side: str = "left", sorter: Optional[np.ndarray] = None) -> np.ndarray:
    return np.searchsorted(a, v, side, sorter)


def zeros(
    shape: Tuple[int, ...],
    *,
    dtype: Optional[TensorDataType] = None,
    device: Optional[TensorDeviceType] = None,
) -> np.ndarray:
    if device is not None and device != TensorDeviceType.CPU:
        raise ValueError("numpy_numeric.zeros only supports CPU device.")
    if dtype is not None:
        dtype = DTYPE_MAP[dtype]
    return np.zeros(shape, dtype=dtype)


def arange(
    start: float,
    end: float,
    step: float,
    *,
    dtype: Optional[TensorDataType] = None,
    device: Optional[TensorDeviceType] = None,
) -> np.ndarray:
    if device is not None and device != TensorDeviceType.CPU:
        raise ValueError("numpy_numeric.arange only supports CPU device.")
    if dtype is not None:
        dtype = DTYPE_MAP[dtype]
    return np.arange(start, end, step, dtype=dtype)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import pathlib
from collections import defaultdict
from typing import Dict

import networkx as nx


def write_dot_graph(G: nx.DiGraph, path: pathlib.Path):
    # NOTE: writing dot files with colons even in labels or other node/edge/graph attributes leads to an
    # error. See https://github.com/networkx/networkx/issues/5962. If `relabel` is True in this function,
    # then the colons (:) will be replaced with (^) symbols.
    relabeled = relabel_graph_for_dot_visualization(G)
    nx.nx_pydot.write_dot(relabeled, str(path))


def get_graph_without_data(G: nx.DiGraph) -> nx.DiGraph:
    """
    Returns the data-less version of the given nx.Digraph. The new graph has no data in nodes and edges.

    :return: nx.DiGraph without data in nodes and edges
    """
    out_graph = nx.DiGraph()
    for node_key in G.nodes(data=False):
        out_graph.add_node(node_key)
    for u, v in G.edges(data=False):
        out_graph.add_edge(u, v)
    return out_graph


def read_dot_graph(path: pathlib.Path) -> nx.DiGraph:
    loaded = nx.DiGraph(nx.nx_pydot.read_dot(str(path)))
    return relabel_graph_for_dot_visualization(loaded, from_reference=True)


RESERVED_CHAR = ":"
REPLACEMENT_CHAR = "^"


def _maybe_escape_colons_in_attrs(data: Dict):
    for attr_name in data:
        attr_val = str(data[attr_name])
        if RESERVED_CHAR in attr_val and not (attr_val[0] == '"' or attr_val[-1] == '"'):
            data[attr_name] = '"' + data[attr_name] + '"'  # escaped colons are allowed


def _unescape_colons_in_attrs_with_colons(data: Dict):
    for attr_name in data:
        attr_val = data[attr_name]
        if RESERVED_CHAR in attr_val and (attr_val[0] == '"' and attr_val[-1] == '"'):
            data[attr_name] = data[attr_name][1:-1]


def _remove_cosmetic_labels(graph: nx.DiGraph):
    for node_name, node_data in graph.nodes(data=True):
        if "label" in node_data:
            label = node_data["label"]
            if node_name == label or '"' + node_name + '"' == label:
                del node_data["label"]


def _add_cosmetic_labels(graph: nx.DiGraph, relabeled_node_mapping: Dict[str, str]):
    for original_name, dot_name in relabeled_node_mapping.items():
        node_data = graph.nodes[dot_name]
        if "label" not in node_data:
            node_data["label"] = '"' + original_name + '"'


def relabel_graph_for_dot_visualization(nx_graph: nx.Graph, from_reference: bool = False) -> nx.DiGraph:
    """
    Relabels NetworkX graph nodes to exclude reserved symbols in keys.
        In case replaced names match for two different nodes, integer index is added to its keys.
        While nodes keys are being updated, visualized nodes names corresponds to the original nodes names.

    :param nx_graph: NetworkX graph to visualize via dot.
    :return: NetworkX graph with reserved symbols in nodes keys replaced.
    """

    nx_graph = copy.deepcopy(nx_graph)

    # .dot format reserves ':' character in node names
    if not from_reference:
        # dumping to disk
        __CHARACTER_REPLACE_FROM = RESERVED_CHAR
        __CHARACTER_REPLACE_TO = REPLACEMENT_CHAR
    else:
        # loading from disk
        __CHARACTER_REPLACE_FROM = REPLACEMENT_CHAR
        __CHARACTER_REPLACE_TO = RESERVED_CHAR

    hits = defaultdict(lambda: 0)
    mapping = {}
    for original_name in nx_graph.nodes():
        if not isinstance(original_name, str):
            continue
        dot_name = original_name.replace(__CHARACTER_REPLACE_FROM, __CHARACTER_REPLACE_TO)
        hits[dot_name] += 1
        if hits[dot_name] > 1:
            dot_name = f"{dot_name}_{hits}"
        if original_name != dot_name:
            mapping[original_name] = dot_name

    relabeled_graph = nx.relabel_nodes(nx_graph, mapping)

    if not from_reference:
        # dumping to disk
        _add_cosmetic_labels(relabeled_graph, mapping)
        for _, node_data in relabeled_graph.nodes(data=True):
            _maybe_escape_colons_in_attrs(node_data)
        for _, _, edge_data in relabeled_graph.edges(data=True):
            _maybe_escape_colons_in_attrs(edge_data)
    else:
        # loading from disk
        _remove_cosmetic_labels(relabeled_graph)
        for _, node_data in relabeled_graph.nodes(data=True):
            _unescape_colons_in_attrs_with_colons(node_data)
        for _, _, edge_data in relabeled_graph.edges(data=True):
            _unescape_colons_in_attrs_with_colons(edge_data)

    return relabeled_graph



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from contextlib import contextmanager

from nncf.common.logging import nncf_logger

DEBUG_LOG_DIR = "./nncf_debug"


def is_debug():
    return nncf_logger.getEffectiveLevel() == logging.DEBUG


def set_debug_log_dir(dir_: str):
    global DEBUG_LOG_DIR
    DEBUG_LOG_DIR = dir_


@contextmanager
def nncf_debug():
    from nncf.common.logging.logger import set_log_level

    set_log_level(logging.DEBUG)
    yield
    set_log_level(logging.INFO)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
@package docstring
This package defines the API for the NNCF compression methods so that the user could
extend the existing algorithms.
"""
from abc import abstractmethod
from typing import Any, Dict, List, Tuple, TypeVar

import torch
from torch import nn

import nncf
from nncf.api.compression import CompressionLoss
from nncf.common.compression import BaseCompressionAlgorithmBuilder
from nncf.common.compression import BaseCompressionAlgorithmController
from nncf.common.graph import NNCFNodeName
from nncf.common.logging import nncf_logger
from nncf.common.scopes import check_scopes_in_graph
from nncf.common.scopes import should_consider_scope
from nncf.config import NNCFConfig
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.layers import NNCF_MODULES_DICT
from nncf.torch.layers import NNCF_WRAPPED_USER_MODULES_DICT
from nncf.torch.model_transformer import PTModelTransformer
from nncf.torch.nncf_network import NNCFNetwork

TModel = TypeVar("TModel")

DOMAIN_CUSTOM_OPS_NAME = "org.openvinotoolkit"


class PTCompressionLoss(nn.Module, CompressionLoss):
    """
    Used to calculate additional loss to be added to the base loss during the
    training process. It uses the model graph to measure variables and activations
    values of the layers during the loss construction. For example, the $L_0$-based
    sparsity algorithm calculates the number of non-zero weights in convolutional
    and fully-connected layers to construct the loss function.
    """

    def calculate(self) -> torch.Tensor:
        """
        Calculates the compression loss value.

        :return: The compression loss value.
        """
        return torch.zeros([])

    def forward(self) -> torch.Tensor:
        """
        Overriding  forward function of the base nn.Module class

        :return: The compression loss value.
        """
        return self.calculate()

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Loads the compression loss state.

        :param state: Output of `get_state()` method.
        """

    def get_state(self) -> None:
        """
        Returns the compression loss state.

        :return: The compression loss state.
        """


class PTCompressionAlgorithmController(BaseCompressionAlgorithmController):
    """
    Serves as a handle to the additional modules, parameters and hooks inserted
    into the original uncompressed model in order to enable algorithm-specific compression.
    Hosts entities that are to be used during the training process, such as compression scheduler and
    compression loss.
    """

    def distributed(self):
        """
        Should be called when distributed training with multiple training processes
        is going to be used (i.e. after the model is wrapped with DistributedDataParallel).
        Any special preparations for the algorithm to properly support distributed training
        should be made inside this function.
        """

    def prepare_for_export(self) -> None:
        # For Torch models no need to call strip_model
        pass


class PTCompressionAlgorithmBuilder(BaseCompressionAlgorithmBuilder):
    """
    Determines which modifications should be made to the original FP32 model in
    order to enable algorithm-specific compression during fine-tuning. Operates
    on an NNCFNetwork object wrapping a target PyTorch model (torch.nn.Module).
    """

    def __init__(self, config: NNCFConfig, should_init: bool = True):
        """
        Arguments:
          `config` - a dictionary that contains parameters of compression method
          `should_init` - if False, trainable parameter initialization will be skipped during building
        """
        super().__init__(config, should_init)
        self.compressed_nncf_module_names = self._nncf_module_types_to_compress()

    def apply_to(self, model: NNCFNetwork) -> NNCFNetwork:
        transformer = PTModelTransformer(model)
        transformation_layout = self.get_transformation_layout(model)
        transformed_model = transformer.transform(transformation_layout)

        if self.should_init:
            self.initialize(transformed_model)

        return transformed_model

    def get_transformation_layout(self, model: NNCFNetwork) -> PTTransformationLayout:
        """
        Applies algorithm-specific modifications to the model. Hooks to be executed during model
        forward operation may be registered using NNCFNetwork command insertion methods. Additional
        compression modules that are expected to be saved along with the network via torch.save should also be
        registered and added to the model here.
        :param model: An instance of NNCFNetwork for the algorithm to be applied to.
        :return: NNCFNetwork with algorithm-specific modifications applied
        """
        check_scopes_in_graph(
            model.nncf.get_original_graph(), self.ignored_scopes, self.target_scopes, self.validate_scopes
        )

        layout = self._get_transformation_layout(model)
        self._handle_frozen_layers(model)
        return layout

    @abstractmethod
    def _build_controller(self, model: TModel) -> PTCompressionAlgorithmController:
        """
        Simple implementation of building controller without setting builder state and loading controller's one.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        :return: The instance of the `BaseCompressionAlgorithmController`.
        """

    def build_controller(self, model: TModel) -> PTCompressionAlgorithmController:
        """
        Builds `PTCompressionAlgorithmController` to handle the additional modules,
        parameters, and hooks inserted into the model to enable algorithm-specific
        compression. Registers the built controller in the model's NNCFNetworkInterface.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        :return: The instance of the `PTCompressionAlgorithmController`.
        """
        ctrl = self._build_controller(model)
        if not isinstance(ctrl, PTCompressionAlgorithmController):
            raise nncf.InternalError(
                "Internal error: builder must create controller inherited from "
                "`PTCompressionAlgorithmController` class"
            )
        ctrl.set_builder_state_with_name(self.name, self.get_state())
        return ctrl

    def _get_state_without_name(self) -> Dict[str, Any]:
        """
        Implementation of get_state that returns state without builder name.

        :return: Returns a dictionary with Python data structures
            (dict, list, tuple, str, int, float, True, False, None) that represents state of the object.
        """
        return {}

    def _load_state_without_name(self, state_without_name: Dict[str, Any]):
        """
        Implementation of load state that takes state without builder name.

        :param state_without_name: Output of `_get_state_without_name()` method.
        """

    def _get_transformation_layout(self, target_model: NNCFNetwork) -> PTTransformationLayout:
        raise NotImplementedError()

    def _handle_frozen_layers(self, target_model: NNCFNetwork):
        scopes_of_frozen_layers = []
        for weighted_node in target_model.nncf.get_weighted_original_graph_nodes():
            should_be_considered = self._should_consider_scope(weighted_node.node_name)
            if not weighted_node.layer_attributes.weight_requires_grad and should_be_considered:
                scopes_of_frozen_layers.append(weighted_node.node_name)
        scopes_to_print = "\n".join(scopes_of_frozen_layers)
        if len(scopes_of_frozen_layers) > 0:
            is_allowed, reason = self._are_frozen_layers_allowed()
            if is_allowed:
                nncf_logger.warning(
                    f"{reason}, compressing them without tuning weights.\nFrozen layers:\n{scopes_to_print}"
                )
            else:
                raise nncf.InternalError(
                    f"{reason}.\n"
                    f"Please unfreeze them or put into the Ignored Scope.\n"
                    f"Frozen Layers:\n"
                    f"{scopes_to_print}"
                )

    def _should_consider_scope(self, node_name: NNCFNodeName) -> bool:
        return should_consider_scope(node_name, self.ignored_scopes, self.target_scopes)

    def _nncf_module_types_to_compress(self) -> List[str]:
        """
        Return list of NNCF module types which should be compressed by specific algorithm.
        As name of algorithm used the value set by decorator @Registry.register() or default one.
        :return: List of names of modules
        """
        filtered_nncf_module_names_list = []
        for module_cls in list(NNCF_MODULES_DICT) + list(NNCF_WRAPPED_USER_MODULES_DICT.values()):
            if self.name not in module_cls.ignored_algorithms:
                filtered_nncf_module_names_list.append(module_cls.__name__)
        return filtered_nncf_module_names_list

    def _are_frozen_layers_allowed(self) -> Tuple[bool, str]:
        algo_name = self.name.replace("_", " ")
        return False, f"Frozen layers are not allowed for {algo_name}"



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.graph.patterns.patterns import GraphPattern as GraphPattern
from nncf.common.graph.patterns.patterns import HWFusedPatternNames as HWFusedPatternNames
from nncf.common.graph.patterns.patterns import IgnoredPatternNames as IgnoredPatternNames
from nncf.common.graph.patterns.patterns import Patterns as Patterns
from nncf.common.graph.patterns.patterns import merge_two_types_of_operations as merge_two_types_of_operations



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from nncf.config.definitions import BOOTSTRAP_NAS_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import EXPERIMENTAL_QUANTIZATION_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import KNOWLEDGE_DISTILLATION_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import MOVEMENT_SPARSITY_ALGO_NAME_IN_CONFIG
from nncf.config.schemata.algo.quantization import QUANTIZATION_SCHEMA
from nncf.config.schemata.basic import ARRAY_OF_NUMBERS
from nncf.config.schemata.basic import ARRAY_OF_STRINGS
from nncf.config.schemata.basic import BOOLEAN
from nncf.config.schemata.basic import NUMBER
from nncf.config.schemata.basic import STRING
from nncf.config.schemata.basic import make_object_or_array_of_objects_schema
from nncf.config.schemata.basic import make_string_or_array_of_strings_schema
from nncf.config.schemata.basic import with_attributes
from nncf.config.schemata.common.compression import BASIC_COMPRESSION_ALGO_SCHEMA
from nncf.config.schemata.common.compression import COMPRESSION_LR_MULTIPLIER_PROPERTY
from nncf.config.schemata.common.initialization import BATCHNORM_ADAPTATION_SCHEMA
from nncf.config.schemata.common.targeting import IGNORED_SCOPES_DESCRIPTION
from nncf.config.schemata.common.targeting import SCOPING_PROPERTIES
from nncf.config.schemata.common.targeting import TARGET_SCOPES_DESCRIPTION

########################################################################################################################
# Experimental Quantization
########################################################################################################################
EXPERIMENTAL_QUANTIZATION_SCHEMA = copy.deepcopy(QUANTIZATION_SCHEMA)
EXPERIMENTAL_QUANTIZATION_SCHEMA["properties"]["algorithm"]["const"] = EXPERIMENTAL_QUANTIZATION_ALGO_NAME_IN_CONFIG

########################################################################################################################
# BootstrapNAS
########################################################################################################################

TRAINING_ALGORITHMS_SCHEMA = {
    "type": "string",
    "enum": ["progressive_shrinking"],
}

ELASTIC_DEPTH_MODE_SCHEMA = {
    "type": "string",
    "enum": ["manual", "auto"],
}

ELASTIC_DEPTH_SCHEMA = {
    "type": "object",
    "properties": {
        "skipped_blocks": {
            "type": "array",
            "items": ARRAY_OF_STRINGS,
            "description": "List of building blocks to be skipped. The block is defined by names of start and end "
            "nodes. The end node is skipped. In contrast, the start node is executed. It produces a "
            "tensor that is bypassed through the skipping nodes until the one after end node. ",
            "examples": [[["start_op_1", "end_op_1"], ["start_op_2", "end_op_2"]]],
        },
        "min_block_size": with_attributes(
            NUMBER,
            description="Defines minimal number of operations in the skipping block. "
            "Option is available for the auto mode only. "
            "Default value is 5",
        ),
        "max_block_size": with_attributes(
            NUMBER,
            description="Defines maximal number of operations in the block. "
            "Option is available for the auto mode only. "
            "Default value is 50",
        ),
        "hw_fused_ops": with_attributes(
            BOOLEAN,
            description="If True, automatic block search will not relate operations, "
            "which are fused on inference, into different blocks for skipping. "
            "True, by default",
        ),
    },
    "additionalProperties": False,
}

ELASTIC_WIDTH_SCHEMA = {
    "type": "object",
    "properties": {
        "min_width": with_attributes(
            NUMBER,
            description="Minimal number of output channels that can be activated for "
            "each layers with elastic width. Default value is 32.",
        ),
        "max_num_widths": with_attributes(
            NUMBER,
            description="Restricts total number of different elastic width values for "
            "each layer. The default value is -1 means that there's no "
            "restrictions.",
        ),
        "width_step": with_attributes(
            NUMBER,
            description="Defines a step size for a generation of the elastic width search "
            "space - the list of all possible width values for each layer. The "
            "generation starts from the number of output channels in the "
            "original model and stops when it reaches whether a "
            "`min_width` width value or number of generated width values "
            "equal to `max_num_widths`",
        ),
        "width_multipliers": with_attributes(
            ARRAY_OF_NUMBERS,
            description="Defines elastic width search space via a list of "
            "multipliers. All possible width values are obtained by "
            "multiplying the original width value with the values in the "
            "given list.",
        ),
        "filter_importance": with_attributes(
            STRING,
            description="The type of filter importance metric. Can be"
            " one of `L1`, `L2`, `geometric_median`, `external`."
            " `L2` by default.",
        ),
        "external_importance_path": with_attributes(
            STRING,
            description="Path to the custom external weight importance (PyTorch tensor) per node "
            "that needs to weight reorder. Valid only when filter_importance "
            "is `external`. The file should be loaded via the torch interface "
            "torch.load(), represented as a dictionary. It maps NNCF node name "
            "to importance tensor with the same shape as the weights in the node "
            "module. For example, node `Model/NNCFLinear[fc1]/linear_0` has a "
            "3x1 linear module with weight [0.2, 0.3, 0.9], and in the dict"
            "{'Model/NNCFLinear[fc1]/linear_0': tensor([0.4, 0.01, 0.2])} represents "
            "the corresponding weight importance.",
        ),
    },
    "additionalProperties": False,
}

ELASTIC_KERNEL_SCHEMA = {
    "type": "object",
    "properties": {
        "max_num_kernels": with_attributes(
            NUMBER,
            description="Restricts the total number of different elastic kernel values for "
            "each layer. The default value is -1 means that there's no "
            "restrictions.",
        ),
    },
    "additionalProperties": False,
}

ELASTICITY_SCHEMA = {
    "type": "object",
    "properties": {
        "depth": ELASTIC_DEPTH_SCHEMA,
        "width": ELASTIC_WIDTH_SCHEMA,
        "kernel": ELASTIC_KERNEL_SCHEMA,
        "available_elasticity_dims": with_attributes(
            ARRAY_OF_STRINGS,
            description="Defines the available elasticity dimension for "
            "sampling subnets. By default, all elastic dimensions "
            "are available - [width, depth, kernel]",
        ),
        "ignored_scopes": with_attributes(
            make_string_or_array_of_strings_schema(),
            description=IGNORED_SCOPES_DESCRIPTION,
        ),
        "target_scopes": with_attributes(
            make_string_or_array_of_strings_schema(),
            description=TARGET_SCOPES_DESCRIPTION,
        ),
    },
    "additionalProperties": False,
}

STAGE_DESCRIPTOR_SCHEMA = {
    "type": "object",
    "properties": {
        "train_dims": with_attributes(
            ARRAY_OF_STRINGS,
            description="Elasticity dimensions that are enabled for subnet sampling,"
            "the rest elastic dimensions are disabled",
        ),
        "epochs": with_attributes(NUMBER, description="Duration of the training stage in epochs"),
        "depth_indicator": with_attributes(
            NUMBER,
            description="Restricts the maximum number of blocks in each "
            "independent group that can be skipped. For example, Resnet50 "
            "has 4 four independent groups, each group consists of a "
            "specific number of Bottleneck layers [3,4,6,3], that "
            "potentially can be skipped. If depth indicator equals to 1,"
            " only the last Bottleneck can be skipped in each group, if it "
            "equals 2 - the last two and etc. This allows to implement "
            "progressive shrinking logic from `Once for all` paper. Default "
            "value is 1.",
        ),
        "width_indicator": with_attributes(
            NUMBER,
            description="Restricts the maximum number of width values in each elastic "
            "layer. For example, some conv2d with elastic width can vary "
            "number of output channels from the following list: [8, 16, 32] "
            "If width indicator is equal to 1, it can only activate the "
            "maximum number of channels - 32. If it equals 2, then the last "
            " two can be selected - 16 or 32, or both of them.",
        ),
        "reorg_weights": with_attributes(
            BOOLEAN,
            description="if True, triggers reorganization of weights in order to have "
            "filters sorted by importance (e.g. by l2 norm) in the "
            "beginning of the stage",
        ),
        "bn_adapt": with_attributes(
            BOOLEAN,
            description="if True, triggers batchnorm adaptation in the beginning of the stage",
        ),
        "init_lr": with_attributes(
            NUMBER,
            description="Initial learning rate for a stage. If specified in the stage "
            " descriptor, it will trigger a reset of the learning rate at "
            "the beginning of the stage.",
        ),
        "epochs_lr": with_attributes(
            NUMBER,
            description="Number of epochs to compute the adjustment of the learning rate.",
        ),
        "sample_rate": with_attributes(
            NUMBER,
            description="Number of iterations to activate the random subnet. Default value is 1.",
        ),
    },
    "description": "Defines a supernet training stage: how many epochs it takes, which elasticities with which "
    "settings are enabled, whether some operation should happen in the beginning",
    "additionalProperties": False,
}
NAS_SCHEDULE_SCHEMA = {
    "type": "object",
    "properties": {
        "list_stage_descriptions": {
            "type": "array",
            "items": STAGE_DESCRIPTOR_SCHEMA,
            "description": "List of parameters per each supernet training stage",
        }
    },
    "additionalProperties": False,
}

LR_SCHEDULE_SCHEMA = {
    "type": "object",
    "properties": {
        "params": {
            "type": "object",
            "properties": {
                "base_lr": with_attributes(
                    NUMBER,
                    description="Defines a global learning rate scheduler."
                    "If these parameters are not set, a stage learning rate scheduler will be used.",
                ),
            },
            "additionalProperties": False,
        }
    },
    "additionalProperties": False,
}

BOOTSTRAP_NAS_TRAINING_SCHEMA = {
    "type": "object",
    "properties": {
        "algorithm": with_attributes(
            TRAINING_ALGORITHMS_SCHEMA,
            description="Defines training strategy for tuning supernet. By default, progressive shrinking",
        ),
        "progressivity_of_elasticity": with_attributes(
            ARRAY_OF_STRINGS,
            description="Defines the order of adding a new elasticity dimension from stage to stage",
            examples=[["width", "depth", "kernel"]],
        ),
        "batchnorm_adaptation": BATCHNORM_ADAPTATION_SCHEMA,
        "schedule": NAS_SCHEDULE_SCHEMA,
        "elasticity": ELASTICITY_SCHEMA,
        "lr_schedule": LR_SCHEDULE_SCHEMA,
        "train_steps": with_attributes(
            NUMBER,
            description="Defines the number of samples used for each training epoch.",
        ),
    },
    "additionalProperties": False,
}

SEARCH_ALGORITHMS_SCHEMA = {
    "type": "string",
    "enum": ["NSGA2", "RNSGA2"],
}

BOOTSTRAP_NAS_SEARCH_SCHEMA = {
    "type": "object",
    "properties": {
        "algorithm": with_attributes(
            SEARCH_ALGORITHMS_SCHEMA,
            description="Defines the search algorithm. Default algorithm is NSGA-II.",
        ),
        "batchnorm_adaptation": BATCHNORM_ADAPTATION_SCHEMA,
        "num_evals": with_attributes(
            NUMBER,
            description="Defines the number of evaluations that will be used by the search algorithm.",
        ),
        "num_constraints": with_attributes(NUMBER, description="Number of constraints in search problem."),
        "population": with_attributes(
            NUMBER,
            description="Defines the population size when using an evolutionary search algorithm.",
        ),
        "crossover_prob": with_attributes(NUMBER, description="Crossover probability used by a genetic algorithm."),
        "crossover_eta": with_attributes(NUMBER, description="Crossover eta."),
        "mutation_eta": with_attributes(NUMBER, description="Mutation eta for genetic algorithm."),
        "mutation_prob": with_attributes(NUMBER, description="Mutation probability for genetic algorithm."),
        "acc_delta": with_attributes(
            NUMBER,
            description="Defines the absolute difference in accuracy that is tolerated "
            "when looking for a subnetwork.",
        ),
        "ref_acc": with_attributes(
            NUMBER,
            description="Defines the reference accuracy from the pre-trained model used "
            "to generate the super-network.",
        ),
        "aspiration_points": with_attributes(
            ARRAY_OF_NUMBERS, description="Information to indicate the preferred parts of the Pareto front"
        ),
        "epsilon": with_attributes(NUMBER, description="epsilon distance of surviving solutions for RNSGA-II."),
        "weights": with_attributes(NUMBER, description="weights used by RNSGA-II."),
        "extreme_points_as_ref_points": with_attributes(
            BOOLEAN, description="Find extreme points and use them as aspiration points."
        ),
        "compression": make_object_or_array_of_objects_schema(
            {"oneOf": [{"$ref": f"#/$defs/{KNOWLEDGE_DISTILLATION_ALGO_NAME_IN_CONFIG}"}]}
        ),
    },
    "additionalProperties": False,
}


BOOTSTRAP_NAS_SCHEMA = {
    "type": "object",
    "properties": {
        "training": BOOTSTRAP_NAS_TRAINING_SCHEMA,
        "search": BOOTSTRAP_NAS_SEARCH_SCHEMA,
    },
    "additionalProperties": False,
}

########################################################################################################################
# Movement Sparsity
########################################################################################################################

MOVEMENT_SPARSE_STRUCTURE_MODE = ["fine", "block", "per_dim"]
MOVEMENT_POWER = 3.0
MOVEMENT_FINAL_IMPORTANCE_THRESHOLD = 0.0
MOVEMENT_ENABLE_STRUCTURED_MASKING = True

MOVEMENT_SPARSE_STRUCTURE_BY_SCOPES_SCHEMA = {
    "type": "object",
    "properties": {
        "mode": with_attributes(
            STRING,
            description="Defines in which mode a supported layer will be sparsified.",
            enum=MOVEMENT_SPARSE_STRUCTURE_MODE,
        ),
        "sparse_factors": with_attributes(
            ARRAY_OF_NUMBERS,
            description='The block shape for weights to sparsify. Required when `mode`="block".',
        ),
        "axis": with_attributes(
            NUMBER,
            description='The dimension for weights to sparsify. Required when `mode`="per_dim".',
        ),
        "target_scopes": with_attributes(
            make_string_or_array_of_strings_schema(),
            description="Model control flow graph node scopes to be considered in this mode.",
        ),
    },
    "additionalProperties": False,
    "required": ["mode", "target_scopes"],
}

MOVEMENT_SCHEDULER_PARAMS_SCHEMA = {
    "type": "object",
    "properties": {
        "warmup_start_epoch": with_attributes(
            NUMBER,
            description="Index of the starting epoch (include) for warmup stage.",
        ),
        "warmup_end_epoch": with_attributes(NUMBER, description="Index of the end epoch (exclude) for warmup stage."),
        "importance_regularization_factor": with_attributes(
            NUMBER,
            description="The regularization factor on weight importance scores. With a larger "
            "positive value, more model weights will be regarded as less important "
            "and thus be sparsified.",
        ),
        "enable_structured_masking": with_attributes(
            BOOLEAN,
            description="Whether to do structured mask resolution after warmup stage. Only "
            "supports structured masking on multi-head self-attention blocks and "
            "feed-forward networks now.",
            default=MOVEMENT_ENABLE_STRUCTURED_MASKING,
        ),
        "power": with_attributes(
            NUMBER,
            description="The power value of polynomial decay for threshold and "
            "regularization factor update during warmup stage.",
            default=MOVEMENT_POWER,
        ),
        "init_importance_threshold": with_attributes(
            NUMBER,
            description="The initial value of importance threshold during warmup stage. If not "
            "specified, this will be automatically decided during training so that "
            "the model is with about 0.1% linear layer sparsity on involved layers at "
            "the beginning of warmup stage.",
        ),
        "final_importance_threshold": with_attributes(
            NUMBER,
            description="The final value of importance threshold during warmup stage.",
            default=MOVEMENT_FINAL_IMPORTANCE_THRESHOLD,
        ),
        "steps_per_epoch": with_attributes(
            NUMBER,
            description="Number of training steps in one epoch, used for proper threshold and "
            "regularization factor updates. Optional if warmup_start_epoch >=1 since "
            "this can be counted in the 1st epoch. Otherwise users have to specify it.",
        ),
    },
    "additionalProperties": False,
    "required": [
        "warmup_start_epoch",
        "warmup_end_epoch",
        "importance_regularization_factor",
    ],
}


MOVEMENT_SPARSITY_SCHEMA = {
    **BASIC_COMPRESSION_ALGO_SCHEMA,
    "properties": {
        "algorithm": {"const": MOVEMENT_SPARSITY_ALGO_NAME_IN_CONFIG},
        "params": MOVEMENT_SCHEDULER_PARAMS_SCHEMA,
        "sparse_structure_by_scopes": {
            "type": "array",
            "items": MOVEMENT_SPARSE_STRUCTURE_BY_SCOPES_SCHEMA,
            "description": "Describes how each supported layer will be sparsified.",
        },
        **SCOPING_PROPERTIES,
        **COMPRESSION_LR_MULTIPLIER_PROPERTY,
    },
    "additionalProperties": False,
}

########################################################################################################################
# All experimental schemas
########################################################################################################################

EXPERIMENTAL_REF_VS_ALGO_SCHEMA = {
    EXPERIMENTAL_QUANTIZATION_ALGO_NAME_IN_CONFIG: EXPERIMENTAL_QUANTIZATION_SCHEMA,
    BOOTSTRAP_NAS_ALGO_NAME_IN_CONFIG: BOOTSTRAP_NAS_SCHEMA,
    MOVEMENT_SPARSITY_ALGO_NAME_IN_CONFIG: MOVEMENT_SPARSITY_SCHEMA,
}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Dict, Optional, Union, cast

from nncf.common.graph.patterns.patterns import GraphPattern
from nncf.common.graph.patterns.patterns import HWFusedPatternNames
from nncf.common.graph.patterns.patterns import IgnoredPatternNames
from nncf.common.graph.patterns.patterns import Patterns
from nncf.common.utils.backend import BackendType
from nncf.parameters import ModelType
from nncf.parameters import TargetDevice

PatternNames = Union[IgnoredPatternNames, HWFusedPatternNames]


class PatternsManager:
    """
    Class provides interface to get hardware or ignored full patterns graph.
    """

    @staticmethod
    def _get_backend_hw_patterns_map(backend: BackendType) -> Dict[HWFusedPatternNames, Callable[[], GraphPattern]]:
        """
        Returns the backend-specific hardware-fused patterns map from the Registry.

        :param backend: BackendType instance.
        :return: Dictionary with the HWFusedPatternNames instance as keys and creator function as a value.
        """
        registry: Dict[HWFusedPatternNames, Callable[[], GraphPattern]] = {}
        if backend == BackendType.ONNX:
            from nncf.onnx.hardware.fused_patterns import ONNX_HW_FUSED_PATTERNS

            registry = ONNX_HW_FUSED_PATTERNS.registry_dict
            return registry
        if backend == BackendType.OPENVINO:
            from nncf.openvino.hardware.fused_patterns import OPENVINO_HW_FUSED_PATTERNS

            registry = OPENVINO_HW_FUSED_PATTERNS.registry_dict
            return registry
        if backend == BackendType.TORCH:
            from nncf.torch.hardware.fused_patterns import PT_HW_FUSED_PATTERNS

            registry = PT_HW_FUSED_PATTERNS.registry_dict
            return registry
        raise ValueError(f"Hardware-fused patterns not implemented for {backend} backend.")

    @staticmethod
    def _get_backend_ignored_patterns_map(
        backend: BackendType,
    ) -> Dict[IgnoredPatternNames, Callable[[], GraphPattern]]:
        """
        Returns the backend-specific ignored patterns map from the Registry.

        :param backend: BackendType instance.
        :return: Dictionary with the HWFusedPatternNames instance as keys and creator function as a value.
        """
        registry: Dict[IgnoredPatternNames, Callable[[], GraphPattern]] = {}
        if backend == BackendType.ONNX:
            from nncf.onnx.quantization.ignored_patterns import ONNX_IGNORED_PATTERNS

            registry = ONNX_IGNORED_PATTERNS.registry_dict
            return registry
        if backend == BackendType.OPENVINO:
            from nncf.openvino.quantization.ignored_patterns import OPENVINO_IGNORED_PATTERNS

            registry = OPENVINO_IGNORED_PATTERNS.registry_dict
            return registry
        if backend == BackendType.TORCH:
            from nncf.torch.quantization.ignored_patterns import PT_IGNORED_PATTERNS

            registry = PT_IGNORED_PATTERNS.registry_dict
            return registry
        raise ValueError(f"Ignored patterns not implemented for {backend} backend.")

    @staticmethod
    def _filter_patterns(
        patterns_to_filter: Dict[PatternNames, Callable[[], GraphPattern]],
        device: TargetDevice,
        model_type: Optional[ModelType] = None,
    ) -> Dict[PatternNames, Callable[[], GraphPattern]]:
        """
        Returns all patterns from patterns_to_filter that are satisfied device and model_type parameters.

        :param patterns_to_filter: Dictionary with the PatternNames instance as keys and creator function as a value.
        :param device: TargetDevice instance.
        :param model_type: ModelType instance.
        :return: Filtered patterns_to_filter.
        """
        filtered_patterns = {}
        for pattern_desc, pattern_creator in patterns_to_filter.items():
            pattern_desc_devices = pattern_desc.value.devices
            pattern_desc_model_types = pattern_desc.value.model_types
            devices_condition = pattern_desc_devices is None or device in pattern_desc_devices
            model_types_condition = pattern_desc_model_types is None or model_type in pattern_desc_model_types
            if devices_condition and model_types_condition:
                filtered_patterns[pattern_desc] = pattern_creator
        return filtered_patterns

    @staticmethod
    def _get_full_pattern_graph(
        backend_patterns_map: Dict[PatternNames, Callable[[], GraphPattern]],
        device: TargetDevice,
        model_type: Optional[ModelType] = None,
    ) -> GraphPattern:
        """
        Filters patterns and returns GraphPattern with registered filtered patterns.

        :param backend_patterns_map: Dictionary with the PatternNames instance as keys and creator function as a value.
        :param device: TargetDevice instance.
        :param model_type: ModelType instance.
        :return: Completed GraphPattern based on the backend, device & model_type.
        """
        filtered_patterns = PatternsManager._filter_patterns(backend_patterns_map, device, model_type)
        patterns = Patterns()
        for pattern_desc, pattern_creator in filtered_patterns.items():
            patterns.register(pattern_creator(), pattern_desc.value.name)
        return patterns.get_full_pattern_graph()

    @staticmethod
    def get_full_hw_pattern_graph(
        backend: BackendType, device: TargetDevice, model_type: Optional[ModelType] = None
    ) -> GraphPattern:
        """
        Returns a GraphPattern containing all registered hardware patterns specifically
        for backend, device, and model_type parameters.

        :param backend: BackendType instance.
        :param device: TargetDevice instance.
        :param model_type: ModelType instance.
        :return: Completed GraphPattern based on the backend, device & model_type.
        """
        backend_patterns_map = cast(
            Dict[PatternNames, Callable[[], GraphPattern]], PatternsManager._get_backend_hw_patterns_map(backend)
        )
        return PatternsManager._get_full_pattern_graph(backend_patterns_map, device, model_type)

    @staticmethod
    def get_full_ignored_pattern_graph(
        backend: BackendType, device: TargetDevice, model_type: Optional[ModelType] = None
    ) -> GraphPattern:
        """
        Returns a GraphPattern containing all registered ignored patterns specifically
        for backend, device, and model_type parameters.

        :param backend: BackendType instance.
        :param device: TargetDevice instance.
        :param model_type: ModelType instance.
        :return: Completed GraphPattern with registered value based on the backend, device & model_type.
        """
        backend_patterns_map = cast(
            Dict[PatternNames, Callable[[], GraphPattern]], PatternsManager._get_backend_ignored_patterns_map(backend)
        )
        return PatternsManager._get_full_pattern_graph(backend_patterns_map, device, model_type)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import List, Tuple

from nncf.common.graph.graph import NNCFNode
from nncf.openvino.graph.layer_attributes import OVLayerAttributes
from nncf.openvino.graph.metatypes.openvino_metatypes import OVConvolutionBackpropDataMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVConvolutionMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVDepthwiseConvolutionMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVGroupConvolutionBackpropDataMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVGroupConvolutionMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVOpMetatype


class OVLayoutElem(Enum):
    """
    Layout elements descriptor for convolutional and linear openvino layers:
        C_IN: Input channels dimension.
        C_OUT: Output channels dimension.
        SPATIAL: Spatial dimension.
        GROUPS: Groups dimention.
    """

    C_IN = "channels_in"
    C_OUT = "channels_out"
    SPATIAL = "spatial"
    GROUPS = "groups"


_CONV_BASE_CONST_LAYOUT = {
    OVConvolutionMetatype: (OVLayoutElem.C_OUT, OVLayoutElem.C_IN),
    OVConvolutionBackpropDataMetatype: (OVLayoutElem.C_IN, OVLayoutElem.C_OUT),
    OVDepthwiseConvolutionMetatype: (OVLayoutElem.GROUPS, OVLayoutElem.C_OUT, OVLayoutElem.C_IN),
    OVGroupConvolutionMetatype: (OVLayoutElem.GROUPS, OVLayoutElem.C_OUT, OVLayoutElem.C_IN),
    OVGroupConvolutionBackpropDataMetatype: (OVLayoutElem.GROUPS, OVLayoutElem.C_IN, OVLayoutElem.C_OUT),
}


def get_conv_weights_layout_from_node(node: NNCFNode) -> List[OVLayoutElem]:
    """
    Calculates weights layout for a target convolution node.

    :param node: Target convolution node.
    :return: Target convolution Node weights layout.
    """
    layer_attributes = node.layer_attributes
    port_id = _get_constant_port_id_from_layer_attributes(layer_attributes)
    return get_conv_weights_layout(
        ov_metatype=node.metatype, weights_shape=layer_attributes.constant_attributes[port_id]["shape"]
    )


def get_linear_weights_layout_from_node(node: NNCFNode) -> List[OVLayoutElem]:
    """
    Calculates weights layout for a target linear node.

    :param node: Target linear node.
    :return: Target linear Node weight layout.
    """
    layer_attributes = node.layer_attributes
    port_id = _get_constant_port_id_from_layer_attributes(layer_attributes)
    constant_layer_attrs = layer_attributes.constant_attributes[port_id]
    return get_linear_weights_layout(
        weights_shape=constant_layer_attrs["shape"],
        transpose=constant_layer_attrs["transpose"],
        port_id=port_id,
    )


def get_conv_weights_layout(ov_metatype: OVOpMetatype, weights_shape: Tuple[int, ...]) -> List[OVLayoutElem]:
    """
    Calculates weights layout for a target convolution node.

    :param ov_metatype: Target convolution node OpenVINO metatype.
    :param weights_shape: Shape of the target convolution node weight.
    :return: Target convolution node weights layout.
    """
    base_layout = _CONV_BASE_CONST_LAYOUT[ov_metatype]
    kernel_size = weights_shape[len(base_layout) :]
    weights_layout = list(base_layout) + [OVLayoutElem.SPATIAL] * len(kernel_size)
    return tuple(weights_layout)


def get_linear_weights_layout(weights_shape: Tuple[int, ...], transpose: bool, port_id: int) -> List[OVLayoutElem]:
    """
    Calculates weights layout for a target linear node.

    :param weights_shape: Shape of the target linear node weight.
    :param port_id: Port id of the target liner node weights.
    :return: Target linear node weight layout.
    """
    weights_layout = [OVLayoutElem.SPATIAL] * (len(weights_shape) - 2)
    if len(weights_shape) > 1:
        if (transpose and port_id == 0) or (not transpose and port_id == 1):
            weights_layout += [OVLayoutElem.C_IN, OVLayoutElem.C_OUT]
        else:
            weights_layout += [OVLayoutElem.C_OUT, OVLayoutElem.C_IN]
    else:
        weights_layout += [OVLayoutElem.C_IN]
    return tuple(weights_layout)


def _get_constant_port_id_from_layer_attributes(layer_attributes: OVLayerAttributes) -> int:
    """
    Returns constant ports id for convolutional and linear ops layer attributes.

    :param layer_attributes: Target convolutional/linear layer op layer attributes.
    :return: Constant port id for the target convolutional/linear model.
    """
    port_ids = list(layer_attributes.constant_attributes.keys())
    assert len(port_ids) == 1
    return port_ids[0]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import numpy as np

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.transformations.command_creation import CommandCreator
from nncf.common.graph.transformations.commands import TargetType
from nncf.openvino.graph.node_utils import get_add_bias_node
from nncf.openvino.graph.transformations.commands import OVBiasCorrectionCommand
from nncf.openvino.graph.transformations.commands import OVBiasInsertionCommand
from nncf.openvino.graph.transformations.commands import OVFQNodeRemovingCommand
from nncf.openvino.graph.transformations.commands import OVMultiplyInsertionCommand
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.graph.transformations.commands import OVWeightUpdateCommand


class OVCommandCreator(CommandCreator):
    """
    Implementation of the `CommandCreator` class for the OpenVINO backend.
    """

    @staticmethod
    def create_command_to_remove_quantizer(quantizer_node: NNCFNode) -> OVFQNodeRemovingCommand:
        target_point = OVTargetPoint(TargetType.LAYER, quantizer_node.node_name, port_id=None)
        return OVFQNodeRemovingCommand(target_point)

    @staticmethod
    def create_command_to_update_bias(
        node_with_bias: NNCFNode, bias_value: np.ndarray, nncf_graph: NNCFGraph
    ) -> OVBiasCorrectionCommand:
        add_node = get_add_bias_node(node_with_bias, nncf_graph)
        const_port_ids = add_node.layer_attributes.get_const_port_ids()
        assert len(const_port_ids) == 1
        bias_port_id = const_port_ids[0]
        target_point = OVTargetPoint(TargetType.LAYER, node_with_bias.node_name, bias_port_id)
        return OVBiasCorrectionCommand(target_point, bias_value)

    @staticmethod
    def create_command_to_update_weight(
        node_with_weight: NNCFNode, weight_value: np.ndarray, weight_port_id: int
    ) -> OVWeightUpdateCommand:
        target_point = OVTargetPoint(TargetType.LAYER, node_with_weight.node_name, weight_port_id)
        return OVWeightUpdateCommand(target_point, weight_value)

    @staticmethod
    def create_command_to_insert_bias(node_without_bias: NNCFNode, bias_value: np.ndarray) -> OVBiasInsertionCommand:
        target_point = OVTargetPoint(TargetType.POST_LAYER_OPERATION, node_without_bias.node_name, 0)
        return OVBiasInsertionCommand(target_point, bias_value)

    @staticmethod
    def multiply_insertion_command(
        source_node: NNCFNode,
        destination_nodes: List[NNCFNode],
        source_out_port: int,
        scale_value: np.ndarray,
        multiply_node_name: str,
    ) -> OVMultiplyInsertionCommand:
        target_point = OVTargetPoint(TargetType.POST_LAYER_OPERATION, source_node.node_name, source_out_port)
        destination_node_names = [d.node_name for d in destination_nodes]
        return OVMultiplyInsertionCommand(target_point, scale_value, destination_node_names, multiply_node_name)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from collections import Counter
from typing import TypeVar

TensorType = TypeVar("TensorType")


class TensorStatistic(ABC):
    """Base class that stores statistic data"""

    TENSOR_STATISTIC_OUTPUT_KEY = "tensor_statistic_output"

    @staticmethod
    @abstractmethod
    def tensor_eq(tensor1: TensorType, tensor2: TensorType, rtol=1e-6) -> bool:
        pass

    @abstractmethod
    def __eq__(self, other):
        pass


class MinMaxTensorStatistic(TensorStatistic):
    MIN_STAT = "min_values"
    MAX_STAT = "max_values"

    def __init__(self, min_values, max_values):
        self.min_values = min_values
        self.max_values = max_values

    def __eq__(self, other: "MinMaxTensorStatistic") -> bool:
        return self.tensor_eq(self.min_values, other.min_values) and self.tensor_eq(self.max_values, other.max_values)


class MeanTensorStatistic(TensorStatistic):
    MEAN_STAT = "mean_values"
    SHAPE_STAT = "shape"

    """
    Base class for the statistics that collects as mean per-axis
    """

    def __init__(self, mean_values, shape):
        """
        :param mean_values: Collected mean per-axis values.
        :param shape: The shape of the collected statistics.
        """
        self.mean_values = mean_values
        self.shape = shape

    def __eq__(self, other: "MeanTensorStatistic") -> bool:
        return self.tensor_eq(self.mean_values, other.mean_values) and self.tensor_eq(self.shape, other.shape)


class MedianMADTensorStatistic(TensorStatistic):
    MEDIAN_VALUES_STAT = "median_values"
    MAD_VALUES_STAT = "mad_values"

    def __init__(self, median_values, mad_values):
        self.median_values = median_values
        self.mad_values = mad_values

    def __eq__(self, other: "MedianMADTensorStatistic") -> bool:
        return self.tensor_eq(self.median_values, other.median_values) and self.tensor_eq(
            self.mad_values, other.mad_values
        )


class PercentileTensorStatistic(TensorStatistic):
    PERCENTILE_VS_VALUE_DICT = "percentile_vs_values_dict"

    def __init__(self, percentile_vs_values_dict):
        self.percentile_vs_values_dict = percentile_vs_values_dict

    def __eq__(self, other: "PercentileTensorStatistic", rtol=1e-9) -> bool:
        if Counter(self.percentile_vs_values_dict.keys()) != Counter(other.percentile_vs_values_dict.keys()):
            return False
        for pct in self.percentile_vs_values_dict:
            if not self.tensor_eq(self.percentile_vs_values_dict[pct], other.percentile_vs_values_dict[pct]):
                return False
        return True


class RawTensorStatistic(TensorStatistic):
    VALUES_STATS = "values"

    """
    Base class for the raw statistics, without any aggregation.
    """

    def __init__(self, values):
        """
        :param values: Collected raw values.
        """
        self.values = values

    def __eq__(self, other: "RawTensorStatistic") -> bool:
        return self.tensor_eq(self.values, other.values)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict

NUMBER = {"type": "number"}
STRING = {"type": "string"}
BOOLEAN = {"type": "boolean"}
ARRAY_OF_NUMBERS = {"type": "array", "items": NUMBER}
ARRAY_OF_STRINGS = {"type": "array", "items": STRING}


def annotated_enum(names_vs_description: Dict[str, str]) -> Dict:
    retval_list = []
    for name, descr in names_vs_description.items():
        retval_list.append({"const": name, "title": name, "description": descr})
    return {"oneOf": retval_list}


def make_string_or_array_of_strings_schema(addtl_dict_entries: Dict = None) -> Dict:
    if addtl_dict_entries is None:
        addtl_dict_entries = {}
    retval = {"type": ["array", "string"], "items": {"type": "string"}}
    retval.update(addtl_dict_entries)
    return retval


def make_object_or_array_of_objects_schema(single_object_schema: Dict = None) -> Dict:
    retval = {
        "oneOf": [
            {
                "title": "single_object_version",
                **single_object_schema,
            },
            {"title": "array_of_objects_version", "type": "array", "items": single_object_schema},
        ]
    }
    return retval


def with_attributes(schema: Dict, **kwargs) -> Dict:
    retval = {**schema, **kwargs}
    return retval



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from collections import deque
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np

from nncf.experimental.tensor.definitions import TensorBackend
from nncf.experimental.tensor.definitions import TensorDataType
from nncf.experimental.tensor.definitions import TensorDeviceType
from nncf.experimental.tensor.definitions import TypeInfo
from nncf.experimental.tensor.functions.dispatcher import dispatch_list
from nncf.experimental.tensor.functions.dispatcher import get_numeric_backend_fn
from nncf.experimental.tensor.functions.dispatcher import tensor_guard
from nncf.experimental.tensor.tensor import Tensor
from nncf.experimental.tensor.tensor import unwrap_tensor_data


@functools.singledispatch
@tensor_guard
def device(a: Tensor) -> TensorDeviceType:
    """
    Return the device of the tensor.

    :param a: The input tensor.
    :return: The device of the tensor.
    """
    return device(a.data)


@functools.singledispatch
@tensor_guard
def backend(a: Tensor) -> TensorBackend:
    """
    Return the backend of the tensor.

    :param a: The input tensor.
    :return: The backend of the tensor.
    """
    return backend(a.data)


@functools.singledispatch
@tensor_guard
def squeeze(a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None) -> Tensor:
    """
    Remove axes of length one from a.

    :param a: The input tensor.
    :param axis: Selects a subset of the entries of length one in the shape.
    :return: The input array, but with all or a subset of the dimensions of length 1 removed.
      This is always a itself or a view into a. Note that if all axes are squeezed,
      the result is a 0d array and not a scalar.
    """
    return Tensor(squeeze(a.data, axis=axis))


@functools.singledispatch
@tensor_guard
def flatten(a: Tensor) -> Tensor:
    """
    Return a copy of the tensor collapsed into one dimension.

    :param a: The input tensor.
    :return: A copy of the input tensor, flattened to one dimension.
    """
    return Tensor(flatten(a.data))


@functools.singledispatch
@tensor_guard
def max(a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: bool = False) -> Tensor:
    """
    Return the maximum of an array or maximum along an axis.

    :param a: The input tensor.
    :param axis: Axis or axes along which to operate. By default, flattened input is used.
    :param keepdim: If this is set to True, the axes which are reduced are left in the result as dimensions with size
        one. With this option, the result will broadcast correctly against the input array. False, by default.
    :return: Maximum of a.
    """
    return Tensor(max(a.data, axis, keepdims))


@functools.singledispatch
@tensor_guard
def min(a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: bool = False) -> Tensor:
    """
    Return the minimum of an array or minimum along an axis.

    :param a: The input tensor.
    :param axis: Axis or axes along which to operate. By default, flattened input is used.
    :param keepdim: If this is set to True, the axes which are reduced are left in the result as dimensions with size
        one. With this option, the result will broadcast correctly against the input array. False, by default.
    :return: Minimum of a.
    """
    return Tensor(min(a.data, axis, keepdims))


@functools.singledispatch
@tensor_guard
def abs(a: Tensor) -> Tensor:
    """
    Calculate the absolute value element-wise.

    :param a: The input tensor.
    :return: A tensor containing the absolute value of each element in x.
    """
    return Tensor(abs(a.data))


@functools.singledispatch
@tensor_guard
def astype(a: Tensor, data_type: TensorDataType) -> Tensor:
    """
    Copy of the tensor, cast to a specified type.

    :param a: The input tensor.
    :param dtype: Type code or data type to which the tensor is cast.

    :return: Copy of the tensor in specified type.
    """
    return Tensor(astype(a.data, data_type))


@functools.singledispatch
@tensor_guard
def dtype(a: Tensor) -> TensorDataType:
    """
    Return data type of the tensor.

    :param a: The input tensor.
    :return: The data type of the tensor.
    """
    return dtype(a.data)


@functools.singledispatch
@tensor_guard
def reshape(a: Tensor, shape: Tuple[int, ...]) -> Tensor:
    """
    Gives a new shape to a tensor without changing its data.

    :param a: Tensor to be reshaped.
    :param shape: The new shape should be compatible with the original shape.
    :return: Reshaped tensor.
    """
    return Tensor(reshape(a.data, shape))


@functools.singledispatch
@tensor_guard
def all(a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None) -> Tensor:
    """
    Test whether all tensor elements along a given axis evaluate to True.

    :param a: The input tensor.
    :param axis: Axis or axes along which a logical AND reduction is performed.
    :return: A new tensor.
    """
    return Tensor(all(a.data, axis=axis))


@functools.singledispatch
@tensor_guard
def allclose(
    a: Tensor, b: Union[Tensor, float], rtol: float = 1e-05, atol: float = 1e-08, equal_nan: bool = False
) -> bool:
    """
    Returns True if two arrays are element-wise equal within a tolerance.

    :param a: The first input tensor.
    :param b: The second input tensor.
    :param rtol: The relative tolerance parameter, defaults to 1e-05.
    :param atol: The absolute tolerance parameter, defaults to 1e-08.
    :param equal_nan: Whether to compare NaN`s as equal. If True,
      NaN`s in a will be considered equal to NaN`s in b in the output array.
      Defaults to False.
    :return: True if the two arrays are equal within the given tolerance, otherwise False.
    """
    return allclose(a.data, unwrap_tensor_data(b), rtol=rtol, atol=atol, equal_nan=equal_nan)


@functools.singledispatch
@tensor_guard
def any(a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None) -> Tensor:
    """
    Test whether any tensor elements along a given axis evaluate to True.

    :param a: The input tensor.
    :param axis: Axis or axes along which a logical OR reduction is performed.
    :return: A new tensor.
    """
    return Tensor(any(a.data, axis))


@functools.singledispatch
@tensor_guard
def count_nonzero(a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None) -> Tensor:
    """
    Counts the number of non-zero values in the tensor input.

    :param a: The tensor for which to count non-zeros.
    :param axis: Axis or tuple of axes along which to count non-zeros.
    :return: Number of non-zero values in the tensor along a given axis.
      Otherwise, the total number of non-zero values in the tensor is returned.
    """
    return Tensor(count_nonzero(a.data, axis))


@functools.singledispatch
@tensor_guard
def isempty(a: Tensor) -> bool:
    """
    Return True if input tensor is empty.

    :param a: The input tensor.
    :return: True if tensor is empty, otherwise False.
    """
    return isempty(a.data)


@functools.singledispatch
@tensor_guard
def isclose(
    a: Tensor, b: Union[Tensor, float], rtol: float = 1e-05, atol: float = 1e-08, equal_nan: bool = False
) -> Tensor:
    """
    Returns a boolean array where two arrays are element-wise equal within a tolerance.

    :param a: The first input tensor.
    :param b: The second input tensor.
    :param rtol: The relative tolerance parameter, defaults to 1e-05.
    :param atol: The absolute tolerance parameter, defaults to 1e-08.
    :param equal_nan: Whether to compare NaN`s as equal. If True,
      NaN`s in a will be considered equal to NaN`s in b in the output array.
      Defaults to False.
    :return: Returns a boolean tensor of where a and b are equal within the given tolerance.
    """
    return Tensor(
        isclose(
            a.data,
            unwrap_tensor_data(b),
            rtol=rtol,
            atol=atol,
            equal_nan=equal_nan,
        )
    )


@functools.singledispatch
@tensor_guard
def maximum(x1: Tensor, x2: Union[Tensor, float]) -> Tensor:
    """
    Element-wise maximum of tensor elements.

    :param x1: The first input tensor.
    :param x2: The second input tensor.
    :return: Output tensor.
    """
    return Tensor(maximum(x1.data, unwrap_tensor_data(x2)))


@functools.singledispatch
@tensor_guard
def minimum(x1: Tensor, x2: Union[Tensor, float]) -> Tensor:
    """
    Element-wise minimum of tensor elements.

    :param x1: The first input tensor.
    :param x2: The second input tensor.
    :return: Output tensor.
    """
    return Tensor(minimum(x1.data, unwrap_tensor_data(x2)))


@functools.singledispatch
@tensor_guard
def ones_like(a: Tensor) -> Tensor:
    """
    Return a tensor of ones with the same shape and type as a given tensor.

    :param a: The shape and data-type of a define these same attributes of the returned tensor.
    :return: Tensor of ones with the same shape and type as a.
    """
    return Tensor(ones_like(a.data))


@functools.singledispatch
@tensor_guard
def where(condition: Tensor, x: Union[Tensor, float], y: Union[Tensor, float]) -> Tensor:
    """
    Return elements chosen from x or y depending on condition.

    :param condition: Where True, yield x, otherwise yield y.
    :param x: Value at indices where condition is True.
    :param y: Value at indices where condition is False.
    :return: A tensor with elements from x where condition is True, and elements from y elsewhere.
    """
    return Tensor(
        where(
            condition.data,
            unwrap_tensor_data(x),
            unwrap_tensor_data(y),
        )
    )


@functools.singledispatch
@tensor_guard
def zeros_like(a: Tensor) -> Tensor:
    """
    Return an tensor of zeros with the same shape and type as a given tensor.

    :param input: The shape and data-type of a define these same attributes of the returned tensor.
    :return: tensor of zeros with the same shape and type as a.
    """
    return Tensor(zeros_like(a.data))


@functools.singledispatch
def stack(x: List[Tensor], axis: int = 0) -> Tensor:
    """
    Stacks a list of Tensors rank-R tensors into one Tensor rank-(R+1) tensor.

    :param x: List of Tensors.
    :param axis: The axis to stack along.
    :return: Stacked Tensor.
    """
    if isinstance(x, (list, deque)):
        return Tensor(dispatch_list(stack, x, axis=axis))
    raise NotImplementedError(f"Function `stack` is not implemented for {type(x)}")


@functools.singledispatch
def concatenate(x: List[Tensor], axis: int = 0) -> Tensor:
    """
    Join a sequence of arrays along an existing axis.

    :param x: The arrays must have the same shape, except in the dimension corresponding to axis.
    :param axis: The axis along which the arrays will be joined. Default is 0.
    :return: The concatenated array.
    """
    if isinstance(x, (list, deque)):
        return Tensor(dispatch_list(concatenate, x, axis=axis))
    raise NotImplementedError(f"Function `concatenate` is not implemented for {type(x)}")


@functools.singledispatch
@tensor_guard
def unstack(a: Tensor, axis: int = 0) -> List[Tensor]:
    """
    Unstack a Tensor into list.

    :param a: Tensor to unstack.
    :param axis: The axis to unstack along.
    :return: List of Tensor.
    """
    res = unstack(a.data, axis=axis)
    return [Tensor(i) for i in res]


@functools.singledispatch
@tensor_guard
def moveaxis(a: Tensor, source: Union[int, Tuple[int, ...]], destination: Union[int, Tuple[int, ...]]) -> Tensor:
    """
    Move axes of an array to new positions.

    :param a: The array whose axes should be reordered.
    :param source: Original positions of the axes to move. These must be unique.
    :param destination: Destination positions for each of the original axes. These must also be unique.
    :return: Array with moved axes.
    """
    return Tensor(moveaxis(a.data, source, destination))


@functools.singledispatch
@tensor_guard
def mean(
    a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: bool = False, dtype: TensorDataType = None
) -> Tensor:
    """
    Compute the arithmetic mean along the specified axis.

    :param a: Array containing numbers whose mean is desired.
    :param axis: Axis or axes along which the means are computed.
    :param keepdims: Destination positions for each of the original axes. These must also be unique.
    :param dtype: Type to use in computing the mean.
    :return: Array with moved axes.
    """
    return Tensor(mean(a.data, axis, keepdims, dtype))


@functools.singledispatch
@tensor_guard
def median(a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: bool = False) -> Tensor:
    """
    Compute the arithmetic median along the specified axis.

    :param a: Array containing numbers whose median is desired.
    :param axis: Axis or axes along which the medians are computed.
    :param keepdims: Destination positions for each of the original axes. These must also be unique.
    :return: Array with moved axes.
    """
    return Tensor(median(a.data, axis, keepdims))


@functools.singledispatch
@tensor_guard
def round(a: Tensor, decimals=0) -> Tensor:
    """
    Evenly round to the given number of decimals.

    :param a: Input data.
    :param decimals: Number of decimal places to round to (default: 0). If decimals is negative,
      it specifies the number of positions to the left of the decimal point.
    :return: An array of the same type as a, containing the rounded values.
    """
    return Tensor(round(a.data, decimals))


@functools.singledispatch
@tensor_guard
def power(a: Tensor, exponent: Union[Tensor, float]) -> Tensor:
    """
    Takes the power of each element in input with exponent and returns a tensor with the result.
    Exponent can be either a single float number or a broadcastable Tensor. In case exponent is
    a brodcastable tensor, the exponent is being broadcasted and the return tensor contains
    the power of each element in input with exponent elementwise.

    :param a: Input data.
    :param exponent: Exponent value.
    :return: The result of the power of each element in input with given exponent.
    """
    return Tensor(power(a.data, unwrap_tensor_data(exponent)))


@functools.singledispatch
@tensor_guard
def quantile(
    a: Tensor,
    q: Union[float, List[float]],
    axis: Optional[Union[int, Tuple[int]]] = None,
    keepdims: Optional[bool] = None,
) -> Tensor:
    """
    Compute the quantile(s) of the data along the specified axis.

    :param a: Given tensor.
    :params q: Quantile or sequence of quantiles to compute, which must be between
        0 and 1 inclusive.
    :param axis: Axis or axes along which the quantiles are computed.
    :param keepdims: If True, the axes which are reduced are left in the result
        as dimensions with size one.
    :return: An tensor with quantiles, the first axis of the result corresponds
        to the quantiles, other axes of the result correspond to the quantiles values.
    """
    return Tensor(quantile(a.data, q, axis, keepdims))


@functools.singledispatch
@tensor_guard
def percentile(
    tensor: Tensor,
    q: Union[float, List[float]],
    axis: Union[int, Tuple[int, ...], List[int]],
    keepdims: bool = False,
) -> Tensor:
    """
    Compute the percentile(s) of the data along the specified axis.

    :param tensor: Given NNCFTensor.
    :params q: percentile or sequence of percentiles to compute, which must be between
        0 and 100 inclusive.
    :param axis: Axis or axes along which the percentiles are computed.
    :param keepdims: If True, the axes which are reduced are left in the result
        as dimensions with size one.
    :returns: The percentile(s) of the tensor elements.
    """
    return Tensor(percentile(tensor.data, q, axis, keepdims))


@functools.singledispatch
@tensor_guard
def _binary_op_nowarn(a: Tensor, b: Union[Tensor, float], operator_fn: Callable) -> Tensor:
    """
    Applies a binary operation with disable warnings.

    :param a: The first tensor.
    :param b: The second tensor.
    :param operator_fn: The binary operation function.
    :return: The result of the binary operation.
    """
    return Tensor(_binary_op_nowarn(a.data, unwrap_tensor_data(b), operator_fn))


@functools.singledispatch
@tensor_guard
def _binary_reverse_op_nowarn(a: Tensor, b: Union[Tensor, float], operator_fn: Callable) -> Tensor:
    """
    Applies a binary reverse operation with disable warnings.

    :param a: The first tensor.
    :param b: The second tensor.
    :param operator_fn: The binary operation function.
    :return: The result of the binary operation.
    """
    return Tensor(_binary_reverse_op_nowarn(a.data, unwrap_tensor_data(b), operator_fn))


@functools.singledispatch
@tensor_guard
def finfo(a: Tensor) -> TypeInfo:
    """
    Returns machine limits for tensor type.

    :param a: Tensor.
    :return: TypeInfo.
    """
    return finfo(a.data)


@functools.singledispatch
@tensor_guard
def clip(a: Tensor, a_min: Union[Tensor, float], a_max: Union[Tensor, float]) -> Tensor:
    """
    Clips all elements in input into the range [ a_min, a_max ]

    :param a: Tensor.
    :param a_min: A lower-bound of the range to be clamped to.
    :param a_max: An upper-bound of the range to be clamped to.
    :return: A clipped tensor with the elements of a, but where values < a_min are replaced with a_min,
        and those > a_max with a_max.
    """
    return Tensor(clip(a.data, unwrap_tensor_data(a_min), unwrap_tensor_data(a_max)))


@functools.singledispatch
@tensor_guard
def as_tensor_like(a: Tensor, data: Any) -> Tensor:
    """
    Converts the data into a tensor with the same data representation and hosted on the same device
    as the given tensor.

    :param a: A tensor for defining the data representation and the host device of the output tensor.
    :param data: Initial data for the tensor. Can be a list, tuple, NumPy ndarray, scalar, and other types.
    :return: A tensor with the same data representation and hosted on the same device as a,
        and which has been initialized with data.
    """
    return Tensor(as_tensor_like(a.data, data))


@functools.singledispatch
@tensor_guard
def item(a: Tensor) -> Union[int, float, bool]:
    """
    Returns the value of this tensor as a standard Python number. This only works for tensors with one element.

    :param a: Tensor.
    :return: The value of this tensor as a standard Python number
    """
    if isinstance(a.data, (int, float, bool)):
        return a.data
    return item(a.data)


@functools.singledispatch
@tensor_guard
def sum(a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: bool = False) -> Tensor:
    """
    Sum of tensor elements over a given axis.

    :param a: The input tensor.
    :param axis: Axis or axes along which a sum is performed. The default, axis=None, will sum all of the elements
        of the input tensor.
    :param keepdims: If this is set to True, the axes which are reduced are left in the result as dimensions
        with size one.
    :return: Returns the sum of all elements in the input tensor in the given axis.
    """
    return Tensor(sum(a.data, axis, keepdims))


@functools.singledispatch
@tensor_guard
def multiply(x1: Tensor, x2: Union[Tensor, float]) -> Tensor:
    """
    Multiply arguments element-wise.

    :param x1: The first input tensor.
    :param x2: The second input tensor or number.
    :return: The product of x1 and x2, element-wise.
    """
    return Tensor(multiply(x1.data, unwrap_tensor_data(x2)))


@functools.singledispatch
@tensor_guard
def var(a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None, keepdims: bool = False, ddof: int = 0) -> Tensor:
    """
    Compute the variance along the specified axis.

    :param a: The input tensor.
    :param axis: Axis or axes along which the variance is computed. The default is to compute the variance
        of the flattened tensor.
    :param keepdims: If this is set to True, the axes which are reduced are left in the result as dimensions
        with size one.
    :param ddof: “Delta Degrees of Freedom”: difference between the sample size and sample degrees of freedom.
        By default ddof is zero.
    :return: A new tensor containing the variance.
    """
    return Tensor(var(a.data, axis, keepdims, ddof))


@functools.singledispatch
@tensor_guard
def size(a: Tensor) -> int:
    """
    Return number of elements in the tensor.

    :param a: The input tensor
    :return: The size of the input tensor.
    """
    return size(a.data)


@functools.singledispatch
@tensor_guard
def matmul(x1: Tensor, x2: Union[Tensor, float]) -> Tensor:
    """
    Matrix multiplication.

    :param x1: The first input tensor.
    :param x2: The second input tensor or number.
    :return: The product of x1 and x2, matmul.
    """
    return Tensor(matmul(x1.data, unwrap_tensor_data(x2)))


@functools.singledispatch
@tensor_guard
def unsqueeze(a: Tensor, axis: Optional[Union[int, Tuple[int, ...]]] = None) -> Tensor:
    """
    Add axes of length one to a.

    :param a: The input tensor.
    :param axis: Selects a subset of the entries of length one in the shape.
    :return: The input array, but with expanded shape with len 1 defined in axis.
    """
    return Tensor(unsqueeze(a.data, axis=axis))


@functools.singledispatch
@tensor_guard
def transpose(a: Tensor, axes: Optional[Tuple[int, ...]] = None) -> Tensor:
    """
    Returns an array with axes transposed.

    :param a: The input tensor.
    :param axes: List of permutations or None.
    :return: A tensor with permuted axes.
    """
    return Tensor(transpose(a.data, axes=axes))


@functools.singledispatch
@tensor_guard
def argsort(a: Tensor, axis: int = -1, descending: bool = False, stable: bool = False) -> Tensor:
    """
    Returns the indices that would sort an array.

    :param a: The input tensor.
    :param axis: Axis along which to sort. The default is -1 (the last axis).
    :param descending: Controls the sorting order (ascending or descending).
    :param stable: If True then the sorting routine becomes stable, preserving the order of equivalent elements.
        If False, the relative order of values which compare equal is not guaranteed. True is slower.
    :return: A tensor of indices that sort a along the specified axis.
    """
    return Tensor(argsort(a.data, axis=axis, descending=descending, stable=stable))


@functools.singledispatch
@tensor_guard
def diag(a: Tensor, k: int = 0) -> Tensor:
    """
    Returns the indices that would sort an array.

    :param a: The input tensor.
    :param k: Diagonal in question. The default is 0. Use k > 0 for diagonals above the main diagonal, and k < 0
        for diagonals below the main diagonal.
    :return: A tensor with the extracted diagonal.
    """
    return Tensor(diag(a.data, k=k))


@functools.singledispatch
@tensor_guard
def logical_or(x1: Tensor, x2: Tensor) -> Tensor:
    """
    Computes the element-wise logical OR of the given input tensors.
    Zeros are treated as False and nonzeros are treated as True.

    :param x1: The input tensor.
    :param x2: The tensor to compute or with.
    :return: Result of elementwise or operation between input_ and other tensor.
    """
    return Tensor(logical_or(x1.data, unwrap_tensor_data(x2)))


@functools.singledispatch
@tensor_guard
def masked_mean(x: Tensor, mask: Tensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims=False) -> Tensor:
    """
    Computes the masked mean of elements across given dimensions of Tensor.

    :param x: Tensor to reduce.
    :param mask: Boolean tensor that have the same shape as x. If an element in mask is True -
        it is skipped during the aggregation.
    :param axis: The dimensions to reduce.
    :param keepdims: If True, the axes which are reduced are left in the result
        as dimensions with size one.
    :return: Reduced Tensor.
    """
    return Tensor(masked_mean(x.data, mask.data, axis, keepdims))


@functools.singledispatch
@tensor_guard
def masked_median(x: Tensor, mask: Tensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims=False) -> Tensor:
    """
    Computes the masked median of elements across given dimensions of Tensor.

    :param x: Tensor to reduce.
    :param axis: The dimensions to reduce.
    :param mask: Boolean tensor that have the same shape as x. If an element in mask is True -
        it is skipped during the aggregation.
    :param keepdims: If True, the axes which are reduced are left in the result
        as dimensions with size one.
    :return: Reduced Tensor.
    """
    return Tensor(masked_median(x.data, mask.data, axis, keepdims))


@functools.singledispatch
@tensor_guard
def clone(a: Tensor) -> Tensor:
    """
    Return a copy of the tensor.

    :param a: The input tensor.
    :return: The copied tensor.
    """
    return Tensor(clone(a.data))


@functools.singledispatch
@tensor_guard
def searchsorted(a: Tensor, v: Tensor, side: str = "left", sorter: Optional[Tensor] = None) -> Tensor:
    """
    Find indices where elements should be inserted to maintain order.

    Find the indices into a sorted tensor a such that, if the corresponding elements in v were inserted
    before the indices, the order of a would be preserved.

    :param a: 1-D tensor. If sorter is None, then it must be sorted in ascending order,
        otherwise sorter must be an array of indices that sort it.
    :param v: Values to insert into a.
    :param side: If 'left', the index of the first suitable location found is given.
        If 'right', return the last such index. Defaults to 'left'.
    :param sorter: Optional array of integer indices that sort array a into ascending order, defaults to None.
    :return: Tensor of insertion points with the same shape as v.
    """
    return Tensor(searchsorted(a.data, v.data, side, unwrap_tensor_data(sorter)))


def zeros(
    shape: Tuple[int, ...],
    *,
    backend: TensorBackend,
    dtype: Optional[TensorDataType] = None,
    device: Optional[TensorDeviceType] = None,
) -> Tensor:
    """
    Return a new array of given shape and type, filled with zeros.

    :param shape: Shape of the new array
    :param backend: The backend type for which the zero tensor is required.
    :param dtype: The data type of the returned tensor, If dtype is not given,
        then the default data type is determined by backend.
    :param device: The device on which the tensor will be allocated, If device is not given,
        then the default device is determined by backend.
    :return: A tensor filled with zeros of the specified shape and data type.
    """
    return Tensor(get_numeric_backend_fn("zeros", backend)(shape, dtype=dtype, device=device))


def arange(
    start: float,
    end: Optional[float] = None,
    step: float = 1,
    *,
    backend: TensorBackend,
    dtype: Optional[TensorDataType] = None,
    device: Optional[TensorDeviceType] = None,
) -> Tensor:
    """
    Returns a tensor with a sequence of numbers in the specified range.

    This function generates a tensor containing a sequence of numbers starting from
    `start` to `end` with a step size of `step`, using the specified backend.

    :param start: The starting value of the sequence.
    :param end: The end value of the sequence (exclusive). If None, the sequence will start at 0 and end at `start`.
    :param step: The step size between each value in the sequence, defaults to 1.
    :param backend: The backend type for which the tensor is required.
    :param dtype: The data type of the returned tensor If dtype is not given, infer the data type
        from the other input arguments.
    :param device: The device on which the tensor will be allocated, If device is not given,
        then the default device is determined by backend.
    :return: A tensor containing the sequence of numbers.
    """
    args = (0, start, step) if end is None else (start, end, step)
    return Tensor(get_numeric_backend_fn("arange", backend)(*args, dtype=dtype, device=device))


def from_numpy(ndarray: np.ndarray, *, backend: TensorBackend) -> Tensor:
    """
    Creates a Tensor from a numpy.ndarray, sharing the memory with the given NumPy array.

    :param ndarray: The numpy.ndarray to share memory with.
    :param backend: The backend type for which the tensor is required.
    :return: A Tensor object that shares memory with the NumPy array.
    """
    if backend == TensorBackend.numpy:
        return Tensor(ndarray)
    return Tensor(get_numeric_backend_fn("from_numpy", backend)(ndarray))



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple

import torch
from torch import Tensor
from torch import nn
from torch.nn.parameter import Parameter

from nncf.common.graph.model_transformer import ModelTransformer
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.torch.extractor import extract_model
from nncf.torch.graph.transformations.commands import ExtraCompressionModuleType
from nncf.torch.graph.transformations.commands import PTBiasCorrectionCommand
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTModelExtractionCommand
from nncf.torch.graph.transformations.commands import PTSharedFnInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.commands import PTWeightUpdateCommand
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.model_graph_manager import update_fused_bias
from nncf.torch.module_operations import UpdateWeight
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.nncf_network import PTInsertionPoint
from nncf.torch.nncf_network import compression_module_type_to_attr_name
from nncf.torch.quantization.external_quantizer import ExternalOpCallHook
from nncf.torch.utils import get_model_device
from nncf.torch.utils import is_multidevice


class PTModelTransformer(ModelTransformer):
    """
    Applies transformations upon PyTorch model.
    """

    def __init__(self, model: NNCFNetwork):
        super().__init__(model)

        device = None
        if not is_multidevice(model):
            device = get_model_device(model)

        self._command_transformation_ordered_pairs = [
            (PTModelExtractionCommand, self._apply_extraction_transformations),
            (PTInsertionCommand, partial(self._apply_insertion_transformations, device=device)),
            (PTSharedFnInsertionCommand, partial(self._apply_shared_nodes_insertion, device=device)),
            (PTBiasCorrectionCommand, self._apply_bias_correction_transformations),
            (PTWeightUpdateCommand, self._apply_weights_update_transformations),
        ]

    def transform(self, transformation_layout: PTTransformationLayout) -> NNCFNetwork:
        transformations = transformation_layout.transformations
        aggregated_transformations = defaultdict(list)
        requires_graph_rebuild = False
        for transformation in transformations:
            aggregated_transformations[transformation.__class__].append(transformation)
            requires_graph_rebuild = requires_graph_rebuild or transformation.requires_graph_rebuild()

        model = self._model
        for transformation_cls, transformation_fn in self._command_transformation_ordered_pairs:
            transformations = aggregated_transformations[transformation_cls]
            if transformations:
                model = transformation_fn(model, transformations)

        if requires_graph_rebuild:
            model.nncf.rebuild_graph()

        return model

    @staticmethod
    def _apply_insertion_transformations(
        model: NNCFNetwork, transformations: List[PTInsertionCommand], device: Optional[torch.device]
    ) -> NNCFNetwork:
        """
        Applies insertion transformations to the model.

        :param model: Model to apply transformations.
        :param transformations: List of the bias correction transformations.
        :param device: Target device for the insertion functions. Applies only to
            functions which are subclassed from torch.nn.Module. Do nothing in case device is None.
        :return: A modified NNCFNetwork.
        """
        node_to_op_address_mapping = model.nncf.get_node_to_op_address_mapping()
        fns_grouped_by_points: Dict[PTInsertionPoint, List[Tuple[Callable, TransformationPriority]]] = defaultdict(list)

        for transformation_command in transformations:
            target_point: PTTargetPoint = transformation_command.target_point
            target_node_name = target_point.target_node_name
            pt_ip = PTInsertionPoint(
                target_type=target_point.target_type,
                op_address=node_to_op_address_mapping[target_node_name],
                input_port_id=target_point.input_port_id,
                replaced_modules=model.nncf.replace_modules,
            )

            fn = transformation_command.fn
            if device is not None and isinstance(fn, torch.nn.Module):
                fn.to(device)

            if model.nncf.replace_modules and target_point.type is TargetType.OPERATION_WITH_WEIGHTS:
                fn = UpdateWeight(fn)
            tup = (fn, transformation_command)
            fns_grouped_by_points[pt_ip].append(tup)

        for pt_ip, fn_list_with_priority in fns_grouped_by_points.items():
            fn_list_with_priority = sorted(fn_list_with_priority, key=lambda x: x[1].priority)
            for fn, command in fn_list_with_priority:
                model.nncf.insert_at_point(pt_ip, fn, command.hooks_group_name)

        return model

    @staticmethod
    def _apply_shared_nodes_insertion(
        model: NNCFNetwork,
        transformations: List[PTSharedFnInsertionCommand],
        device: Optional[torch.device],
    ) -> NNCFNetwork:
        """
        Applies insertion of PTSharedFnInsertionCommand commands. For each command method inserts
        a torch module to the NNCFNetwork and inserts call hooks for each command target points.

        :param model: Model to apply transformations.
        :param transformations: List of the bias correction transformations.
        :param device: Target device for the insertion functions. Applies only to
            functions which are subclassed from torch.nn.Module. Do nothing in case device is None.
        :return: A modified NNCFNetwork.
        """
        compression_type_vs_transformations = defaultdict(list)
        for transformation in transformations:
            compression_type_vs_transformations[transformation.compression_module_type].append(transformation)

        for compression_module_type, transformations in compression_type_vs_transformations.items():
            model = PTModelTransformer._apply_shared_node_insertion_with_compression_type(
                model, transformations, device, compression_module_type
            )
        return model

    @staticmethod
    def _apply_shared_node_insertion_with_compression_type(
        model: NNCFNetwork,
        transformations: List[PTSharedFnInsertionCommand],
        device: Optional[torch.device],
        compression_module_type: ExtraCompressionModuleType,
    ):
        """
        Does _apply_shared_nodes_insertion with specified compression model type which will be
        used for each transformation command.

        :param model: Model to apply transformations.
        :param transformations: List of the bias correction transformations.
        :param device: Target device for the insertion functions. Applies only to
            functions which are subclassed from torch.nn.Module. Do nothing in case device is None.
        :param compression_module_type: Common compression module type for all commands.
        :return: A modified NNCFNetwork.
        """
        if not model.nncf.is_compression_module_registered(compression_module_type):
            model.nncf.register_compression_module_type(compression_module_type)

        insertion_commands: List[PTInsertionCommand] = []

        for shared_command in transformations:
            fn = shared_command.fn
            if device is not None:
                fn.to(device)

            model.nncf.add_compression_module(shared_command.op_name, fn, compression_module_type)

            for target_point in shared_command.target_points:
                fn = ExternalOpCallHook(
                    compression_module_type_to_attr_name(compression_module_type), shared_command.op_name
                )
                insertion_commands.append(
                    PTInsertionCommand(
                        target_point,
                        fn,
                        priority=shared_command.priority,
                        hooks_group_name=shared_command.hooks_group_name,
                    )
                )

        return PTModelTransformer._apply_insertion_transformations(model, insertion_commands, device)

    @staticmethod
    def _apply_extraction_transformations(
        model: NNCFNetwork, transformations: List[PTModelExtractionCommand]
    ) -> nn.Sequential:
        """
        Extracts copy of sub-modules.

        :param model: Model to apply transformations.
        :param transformation: Model extraction transformation.
        :return: Extracted sub-modules.
        """
        transformation = transformations[-1]
        return extract_model(model, transformation.input_node_names, transformation.output_node_names)

    @staticmethod
    def _apply_bias_correction_transformations(
        model: NNCFNetwork, transformations: List[PTBiasCorrectionCommand]
    ) -> NNCFNetwork:
        """
        Applies bias correction transformations on the model.

        :param model: Model to apply transformations.
        :param transformations: List of the bias correction transformations.
        :return: Model with corrected bias.
        """
        for transformation in transformations:
            update_fused_bias(
                target_node_name=transformation.target_point.target_node_name,
                new_bias=transformation.bias_value,
                model=model,
            )
        return model

    @staticmethod
    def _apply_weights_update_transformations(
        model: NNCFNetwork, transformations: List[PTWeightUpdateCommand]
    ) -> NNCFNetwork:
        """
        Applies weight update transformations on the model.

        :param model: Model to apply transformations.
        :param transformations: List of the weight update transformations.
        :return: Model with updated weights.
        """
        for transformation in transformations:
            update_parameter(transformation.target_point.target_node_name, "weight", transformation.weight_value, model)
        return model


def update_parameter(target_node_name: str, parameter_name: str, new_value: Tensor, model: NNCFNetwork) -> None:
    """
    Update parameter for target module.

    :param target_node_name: The target node name.
    :param parameter_name: The name of the parameter to update.
    :param new_value: New parameter value.
    :param model: The model.
    """
    module = model.nncf.get_containing_module(target_node_name)
    parameter: Parameter = getattr(module, parameter_name)
    parameter.data = new_value
